from __future__ import absolute_import
from __future__ import print_function

import socket
import time
import rosnode
from . import rospy_safe as rospy

# We wrap rosnode function into safeguards for socket error
# since the master seems to be quite sensitive to Network health.
# TODO : refine which ones need which exception handling...

retry_timeout = 1  # number of seconds to wait before retrying


def get_api_uri(master, caller_id):
    res = None
    while res is None:
        try:
            res = rosnode.get_api_uri(master, caller_id)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr(
                "Pyros : got socket error calling get_api_uri({master}, {caller_id}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


from __future__ import absolute_import

from collections import Counter

from .api import rosnode_safe as rosnode

from .api import rospy_safe as rospy
from .message_conversion import get_msg, get_msg_dict


def get_topic_msg(topic):
    return get_msg(topic.rostype)


def get_topic_msg_dict(topic):
    return get_msg_dict(topic.rostype)


class PoolParam(object):
    """
    This is a pool, supported by the param server in ROS
    It optimizes transient (pub/sub, etc.) creation and reuse, and should play nicely with multiple pyros processes.

    The param name for one interface is :
    - /pyros to gather all pyros instances
    - /<node_name> to gather interfaces by pyros node instances
    - /<if_description> to gather different kind of interfaces (pubs, subs, etc.)
    - /<if_name>

    The param state follows this :
    INTERFACE ON -> param created and set to True -> means currently active
      => if exists and set to true we can ignore the "added interface" from a potential delayed diff feedback
    INTERFACE OFF -> param set to False -> means was active before, currently deactivated
      => if exists and set to false we can ignore the "removed interface" from a potential delayed diff feedback
    param not set means interface was never activated.

    """

    # need to wait for ROS to start before we can get a name...
    @property
    def param_namespace(self):
        return '/pyros' + rospy.get_name() + '/' + self.topic_descr

    def __init__(self, topic_class, topic_descr):
        # :topic_class rospy.Publisher or rospy.Subscriber
        self.topic_class = topic_class
        # :topic_descr "publishers" or "subscribers"
        self.topic_descr = topic_descr
        # ROS didnt start yet we cant write a param.
        # rospy.set_param(self.param_namespace, {})

        # setting up the pool for this instance
        self.topics = {}
        self.topics_count = Counter()

    def acquire(self, topic_name, topic_type, *args, **kwargs):
        """
        Creating a publisher (if needed) and adding it to the pub instance count.
        Otherwise will reuse an existing publisher for the same topic.
        :return: the ros publisher
        """
        # NOTE : the node must be initialized before reaching this
        topic_name = rospy.resolve_name(topic_name)
        if topic_name not in self.topics:
            # Asserting this topic interface is also not registered on ROS param server
            # TODO : fix this. This currently can assert because cleanup is not happening when it should (check for shutting_down argument to update()).
            #assert (not rospy.get_param(self.param_namespace + topic_name, False))
            # build a new instance only if needed
            self.topics[topic_name] = self.topic_class(topic_name, topic_type, *args, **kwargs)

            # Advertising ROS system wide, which topic are interfaced with this process
            #  We need this to be atomic to avoid race conditions
            rospy.set_param(self.param_namespace + topic_name, True)

        # assert topic type (data_class) didn't change in the meantime (ROS doesnt support it anyway)
        assert (topic_type == self.topics[topic_name].data_class)
        # We count artificial instances here (rospy uses same socket internally).
        # this helps using only one publisher in the interface.
        self.topics_count[topic_name] += 1
        # but that data has meaning only in this process. other processes will see only one connection.
        return self.topics[topic_name]

    def release(self, tpc):
        """
        Removing a topic and substracting it from the list.
        :return: None
        """
        if tpc.name in self.topics_count:
            self.topics_count[tpc.name] -= 1
            if self.topics_count[tpc.name] == 0:
                # Advertising ROS system wide, which topic are interfaced with this process
                #  We need this to be atomic to avoid race conditions
                rospy.set_param(self.param_namespace + tpc.name, False)

                self.topics.pop(tpc.name)
                # TODO: keep it around until GC ??
                # Be aware of https://github.com/ros/ros_comm/issues/111
                tpc.unregister()

    def get_impl_ref_count(self, name):
        """
        Returns the reference counter for this topic interface
        :param name: name of the topic
        :return: reference count of topic interface implementation
        """

        res = 0
        if name in self.topics:
            res = self.topics[name].impl.ref_count
            # TODO : how about topic counter ?
        #print("get_impl_ref_count({name}) => {res}".format(**locals()))
        return res

    def get_impl_connections(self, name):
        """
        Returns the connections for this topic interface
        :param name: name of the topic
        :return: connections of topic interface implementation
        """

        res = []
        if name in self.topics:
            # get usable connection representation as tuple
            res = self.topics[name].impl.get_stats_info()
            # TODO : how about topic counter ?
        #print("get_impl_connections({name}) => {res}".format(**locals()))
        return res

    def get_all_interfaces(self, ros_node=None):
        # Parameters will interpret / as a sub mapping collection
        # We need to flatten it to retrieve topic names...
        def flatten_dict(d):
            def expand(key, value):
                if isinstance(value, dict):
                    return [(key + '/' + k, v) for k, v in flatten_dict(value).items()]
                else:
                    return [(key, value)]

            items = [item for k, v in d.items() for item in expand(k, v)]

            return dict(items)

        # Inspect params to find who also interface this publisher
        # TODO : use hte param internal interface instead of doing a request here...
        # TODO : or maybe NOT ! The synchronicity of this call is important to know the current state of the interface
        # TODO : getting it from cache suddenly break this mandatory synchronous access...
        pyros_if = rospy.get_param('/pyros', {})
        if_map = {}
        for k, v in pyros_if.iteritems():
            # we need to reconstruct the slashes, lost when storing as params...
            if_map["/" + k] = {
                'uri': rosnode.get_api_uri(rospy.get_master(), "/" + k)[2],
                self.topic_descr: {"/" + tn: tv for tn, tv in flatten_dict(v.get(self.topic_descr, {})).iteritems()}
            }
        return if_map

    def __del__(self):
        rospy.delete_param('/pyros' + rospy.get_name() + '/' + self.topic_descr)




###
# Generic
###
# Settings to pass to pyros node to interface with another system

TOPICS = []
SERVICES = []
PARAMS = []

###
# ROS specific
###
ROS_USE_CONNECTION_CACHE = False
ROS_CONNECTION_CACHE_LIST_TOPIC = "/rocon/connection_cache/list"
ROS_CONNECTION_CACHE_DIFF_TOPIC = "/rocon/connection_cache/diff"

###
# Mock specific
###
# MOCK_



from __future__ import absolute_import

import time

import roslib
import rospy

from importlib import import_module
from collections import deque, OrderedDict


from .message_conversion import get_msg, get_msg_dict, populate_instance, extract_values, FieldTypeMismatchException
from .poolparam import PoolParam
from .topicbase import TopicBase


class SubscriberBackTimeout(Exception):
    pass


class SubscriberBack(TopicBase):
    """
    PublisherBack is the class handling conversion from Python to ROS Publisher
    Requirement : Only one publisherBack per actual ROS Topic.
    Since we connect to an already existing ros topic, our number of connections should never drop under 1
    """

    # We need some kind of instance count here since system state returns only one node instance
    # as publisher for multiple publisher in this process.
    #
    # This is used in ros_interface update for added pubs/subs to determine
    # if we previously started publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests
    #
    # This is also used in ros_interface update for removed pubs/subs to determine
    # if we previously stopped publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests

    pool = PoolParam(rospy.Publisher, "publishers")

    def __init__(self, topic_name, topic_type):

        super(SubscriberBack, self).__init__(topic_name, topic_type)
        # Is 1 a good choice ? # TODO : check which value is best here...

        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Adding publisher interface {name} {typename}".format(
                name=self.name, typename=self.rostype))

        self.topic = self.pool.acquire(self.name, self.rostype, queue_size=1)
        # CAREFUL ROS publisher doesnt guarantee messages to be delivered
        # stream-like design spec -> loss is acceptable.

        # Make sure we get at least one connection before returning
        start = time.time()
        timeout = 1
        while time.time() - start < timeout and len(self.pool.get_impl_connections(self.name)) < 1:
            # print("publishers connected : {0}".format(self.pool.get_impl_connections(self.name)))
            time.sleep(0.2)

        if self.pool.get_impl_connections(self.name) < 1:
            raise SubscriberBackTimeout

    def cleanup(self):
        """
        Launched when we want to whithhold this interface instance
        :return:
        """
        # TODO : should we do this in del method instead ? to allow reuse until garbage collection actually happens...
        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Removing publisher interface {name} {typename}".format(
                name=self.name, typename=self.rostype))

        self.pool.release(self.topic)

        super(SubscriberBack, self).cleanup()

    def asdict(self):
        """
        Here we provide a dictionary suitable for a representation of the Topic instance
        the main point here is to make it possible to transfer this to remote processes.
        We are not interested in pickleing the whole class with Subscriber and Publisher
        :return:
        """
        d = super(SubscriberBack, self). asdict()
        d['subscribers'] = self.topic.impl.get_stats_info()
        return d

    def publish(self, msg_content):
        """
        Publishes a message to the topic
        :return the actual message sent if one was sent, None if message couldn't be sent.
        """
        # enforcing correct type to make send / receive symmetric and API less magical
        # Doing message conversion visibly in code before sending into the black magic tunnel sounds like a good idea
        try:
            msg = self.rostype()
            populate_instance(msg_content, msg)
            if isinstance(msg, self.rostype):
                self.topic.publish(msg)  # This should return False if publisher not fully setup yet
                return msg  # because the return spec of rospy's publish is not consistent
        except FieldTypeMismatchException as e:
            rospy.logerr("[{name}] : field type mismatch {e}".format(name=__name__, e=e))
            raise
            # TODO : reraise a topic exception ?
        return None




from __future__ import absolute_import, print_function

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..')))

import nose
import unittest
from nose.tools import assert_true, assert_false, assert_raises


from pyros.rosinterface.mock import extract_values, populate_instance, FieldTypeMismatchException, NonexistentFieldException, StatusMsg
from pyros.rosinterface.mock import statusecho_topic, MockSystem


class TestMockTopic(unittest.TestCase):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        cls.system = MockSystem()

    @classmethod
    def teardown_class(cls):
        pass

    def test_echo_status(self):
        msg = populate_instance({
            'error': 'BAD_ERROR',
            'code': 7,
            'message': 'A bad Error happened'
        }, StatusMsg('ERROR', 42, 'details'))
        assert_true(isinstance(msg, StatusMsg))
        echo_topic_pub = self.system.create_publisher('random_topic', statusecho_topic)
        echo_topic_sub = self.system.create_subscriber('random_topic', statusecho_topic)
        assert_true(echo_topic_pub.publish(msg))
        recv = echo_topic_sub.get()
        print(recv)
        assert_true(recv.error == 'BAD_ERROR')
        assert_true(recv.code == 7)
        assert_true(recv.message == 'A bad Error happened')


if __name__ == '__main__':

    nose.runmodule()



#!/usr/bin/env python
from __future__ import absolute_import

import os
import sys
import pickle

# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
import time

# current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
# # if not current_path in sys.path:
# sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# setting up logging from this test
import logging.config
# Setting up logging for this test
# TODO : solve multiprocess problems... this seems not enough to get on console...
logging.config.dictConfig(
    {
        'version': 1,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(name)s:%(message)s'
            },
        },
        'handlers': {
            'null': {
                'level': 'DEBUG',
                'class': 'logging.NullHandler',
            },
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
        },
        'loggers': {
            'pyros_config': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
            'pyros_setup': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'pyros': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'pyros.rosinterface': {
                'handlers': ['console'],
                'level': 'DEBUG',
            },
            'pyros.ros_interface': {
                'handlers': ['console'],
                'level': 'DEBUG',
            }
        }
    }
)


# Unit test import (  will emulate ROS setup if needed )
import nose

from pyros.baseinterface import DiffTuple
from pyros.rosinterface import RosServiceIfPool
from pyros.rosinterface.connection_cache_utils import connection_cache_proxy_create, connection_cache_marshall


import rospy
import roslaunch
import rosnode
from std_msgs.msg import String, Empty
from std_srvs.srv import Empty as EmptySrv, Trigger

from pyros.rosinterface.rostests import Timeout


# useful test tools
from pyros_utils import rostest_nose
import unittest


# This should have the same effect as the <name>.test file for rostest.
# Should be used only by nose ( or other python test tool )
# CAREFUL with comments, copy paste mistake are real...
# CAREFUL dont use assertFalse -> easy to miss when reviewing code
def setup_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_setup_module()

    # Providing services require us to be a node
    rospy.init_node('test_service_if_pool', argv=None, disable_signals=True)
    # CAREFUL : rospy.init_node should be done only once per PROCESS
    # Here we enforce TEST RUN 1<->1 MODULE 1<->1 PROCESS. ROStest style.


def teardown_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_teardown_module()


# Very basic echo service implementation for tests
def srv_cb(req):
    return req.request


@nose.tools.nottest
class TestRosServiceIfPool(unittest.TestCase):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """

    launch = None
    # test node process not setup by default (rostest dont need it here)
    empty_srv_process = None
    trigger_srv_process = None

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        # Start roslaunch
        TestRosServiceIfPool.launch = roslaunch.scriptapi.ROSLaunch()
        TestRosServiceIfPool.launch.start()

        # start required nodes - needs to match the content of *.test files for rostest to match
        global empty_srv_process, trigger_srv_process
        empty_srv_node = roslaunch.core.Node('pyros_test', 'emptyService.py', name='empty_service')
        trigger_srv_node = roslaunch.core.Node('pyros_test', 'triggerService.py', name='trigger_service')
        TestRosServiceIfPool.empty_srv_process = TestRosServiceIfPool.launch.launch(empty_srv_node)
        TestRosServiceIfPool.trigger_srv_process = TestRosServiceIfPool.launch.launch(trigger_srv_node)

    @classmethod
    def teardown_class(cls):
        # ensuring all process are finished
        if TestRosServiceIfPool.empty_srv_process is not None:
            TestRosServiceIfPool.empty_srv_process.stop()
        if TestRosServiceIfPool.trigger_srv_process is not None:
            TestRosServiceIfPool.trigger_srv_process.stop()

    # EXPOSE SERVICES + UPDATE Interface

    def test_service_appear_expose_update(self):
        """
        Test service exposing functionality for a service which already exists in
        the ros environment. Simple Normal usecase
        Sequence : APPEAR -> EXPOSE -> UPDATE
        :return:
        """

        servicename = '/test/empsrv'
        dt = self.service_if_pool.expose_services([servicename])
        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # service backend has not been created since the update didn't run yet
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        # NOTE : We need to wait to make sure the tests nodes are started...
        with Timeout(5) as t:
            while not t.timed_out and servicename not in dt.added:
                services, service_types = self.get_system_state()
                dt = self.service_if_pool.update(services, service_types)
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        # every exposed service should remain in the list of args ( in case regex match another service )
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # make sure the service backend has been created
        self.assertTrue(servicename in self.service_if_pool.services.keys())

        # cleaning up
        self.service_if_pool.expose_services([])
        # service should not be in the list of args any longer
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # service backend has been removed
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

    def test_service_appear_update_expose(self):
        """
        Test service exposing functionality for a service which already exists in
        the ros environment. Normal usecase
        Sequence : (UPDATE?) -> APPEAR -> UPDATE -> EXPOSE (-> UPDATE?)
        :return:
        """
        servicename = '/test/absentsrv1'
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())
        # First update should not change state
        services, service_types = self.get_system_state()
        dt = self.service_if_pool.update(services, service_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        # create the service and then try exposing the service again, simulating
        # it coming online before expose call.
        nonexistent_srv = rospy.Service(servicename, EmptySrv, srv_cb)
        try:
            with Timeout(5) as t:
                while not t.timed_out and nonexistent_srv.resolved_name not in self.service_if_pool.services_available:
                    services, service_types = self.get_system_state()
                    dt = self.service_if_pool.update(services, service_types)
                    self.assertEqual(dt.added, [])  # nothing added (not exposed yet)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            # every added service should be in the list of args
            self.assertTrue(servicename not in self.service_if_pool.services_args)
            # the backend should not have been created
            self.assertTrue(servicename not in self.service_if_pool.services.keys())

            # here we are sure the interface knows the service is available
            # it will be exposed right now
            dt = self.service_if_pool.expose_services([servicename])
            self.assertTrue(servicename in dt.added)  # servicename added
            self.assertEqual(dt.removed, [])  # nothing removed

            # every exposed service should remain in the list of args ( in case regex match another service )
            self.assertTrue(servicename in self.service_if_pool.services_args)
            # make sure the service backend has been created
            self.assertTrue(servicename in self.service_if_pool.services.keys())
        finally:
            nonexistent_srv.shutdown('testing complete')

    def test_service_expose_appear_update(self):
        """
        Test basic service adding functionality for a service which does not yet exist
        in the ros environment ( + corner cases )
        Sequence : (UPDATE? ->) -> EXPOSE -> (UPDATE? ->) APPEAR -> UPDATE
        :return:
        """
        servicename = '/test/absentsrv1'
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())
        # First update should not change state
        services, service_types = self.get_system_state()
        dt = self.service_if_pool.update(services, service_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        self.service_if_pool.expose_services([servicename])
        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())
        services, service_types = self.get_system_state()
        dt = self.service_if_pool.update(services, service_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # make sure the service is STILL in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # make sure the service backend has STILL not been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        # create the service and then try updating again, simulating
        # it coming online after expose call.
        nonexistent_srv = rospy.Service(servicename, EmptySrv, srv_cb)
        try:
            with Timeout(5) as t:
                dt = DiffTuple([], [])
                while not t.timed_out and nonexistent_srv.resolved_name not in dt.added:
                    services, service_types = self.get_system_state()
                    dt = self.service_if_pool.update(services, service_types)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(nonexistent_srv.resolved_name in dt.added)  # nonexistent_srv added
            # every exposed service should remain in the list of args ( in case regex match another service )
            self.assertTrue(servicename in self.service_if_pool.services_args)
            # make sure the service backend has been created
            self.assertTrue(servicename in self.service_if_pool.services.keys())
        finally:
            nonexistent_srv.shutdown('testing complete')

    def test_service_withhold_update_disappear(self):
        """
        Test service witholding functionality for a service which doesnt exists anymore in
        the ros environment. Normal usecase.
        Sequence : (-> UPDATE ?) -> WITHHOLD -> UPDATE -> DISAPPEAR (-> UPDATE ?)
        :return:
        """
        servicename = '/test/absentsrv1'
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        self.service_if_pool.expose_services([servicename])
        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # service backend has NOT been created yet
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        dt = DiffTuple([], [])
        # create the service and then try updating again, simulating
        # it coming online after expose call.
        nonexistent_srv = rospy.Service(servicename, EmptySrv, srv_cb)
        try:
            # wait here until service actually appear in cache proxy
            with Timeout(5) as t:
                while not t.timed_out and nonexistent_srv.resolved_name not in dt.added:
                    services, service_types = self.get_system_state()
                    dt = self.service_if_pool.update(services, service_types)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(nonexistent_srv.resolved_name in dt.added)  # nonexistent_srv added
            self.assertEqual(dt.removed, [])  # nothing removed

            # every withhold service should STILL be in the list of args
            self.assertTrue(servicename in self.service_if_pool.services_args)
            # service backend has been created
            self.assertTrue(servicename in self.service_if_pool.services.keys())

            dt = self.service_if_pool.expose_services([])
            self.assertEqual(dt.added, [])  # nothing added
            self.assertTrue(nonexistent_srv.resolved_name in dt.removed)  # nonexistent_srv removed
            # every withhold service should NOT be in the list of args
            self.assertTrue(servicename not in self.service_if_pool.services_args)
            # service backend should be GONE
            self.assertTrue(servicename not in self.service_if_pool.services.keys())

            services, service_types = self.get_system_state()
            dt = self.service_if_pool.update(services, service_types)
            self.assertEqual(dt.added, [])  # nothing added
            self.assertEqual(dt.removed, [])  # nothing removed
            # every withhold service should STILL NOT be in the list of args
            self.assertTrue(servicename not in self.service_if_pool.services_args)
            # service backend should be GONE
            self.assertTrue(servicename not in self.service_if_pool.services.keys())
        finally:
            nonexistent_srv.shutdown('testing disappearing service')

        services, service_types = self.get_system_state()
        dt = self.service_if_pool.update(services, service_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nonexistent_srv already removed
        # every withhold service should STILL NOT be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # service backend should be GONE
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

    def test_service_disappear_update_withhold(self):
        """
        Test service exposing functionality for a service which already exists in
        the ros environment. Normal usecase
        Sequence : (UPDATE? ->) DISAPPEAR -> UPDATE -> WITHHOLD (-> UPDATE ?)
        :return:
        """
        servicename = '/test/absentsrv1'
        # service should not be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        self.service_if_pool.expose_services([servicename])
        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # service backend has NOT been created yet
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        dt = DiffTuple([], [])
        # create the service and then try updating again, simulating
        # it coming online after expose call.
        nonexistent_srv = rospy.Service(servicename, EmptySrv, srv_cb)
        try:
            # wait here until service actually appear in cache proxy
            with Timeout(5) as t:
                while not t.timed_out and nonexistent_srv.resolved_name not in dt.added:
                    services, service_types = self.get_system_state()
                    dt = self.service_if_pool.update(services, service_types)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(nonexistent_srv.resolved_name in dt.added)  # nonexistent added
            self.assertEqual(dt.removed, [])  # nothing removed

            # service should be in the list of args
            self.assertTrue(servicename in self.service_if_pool.services_args)
            # the backend should have been created
            self.assertTrue(servicename in self.service_if_pool.services.keys())

        # up to here possible sequences should have been already tested by previous tests
        # Now comes our actual disappearance / withholding test
        finally:
            nonexistent_srv.shutdown('testing disappearing service')

        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # the backend should STILL be there
        self.assertTrue(servicename in self.service_if_pool.services.keys())
        # Note the service implementation should take care of possible errors in this case

        # wait here until service actually disappear from cache proxy
        with Timeout(5) as t:
            while not t.timed_out and nonexistent_srv.resolved_name not in dt.removed:
                services, service_types = self.get_system_state()
                dt = self.service_if_pool.update(services, service_types)
                self.assertEqual(dt.added, [])  # nothing added
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(nonexistent_srv.resolved_name in dt.removed)  # nonexistent_srv removed
        # every exposed service should remain in the list of args ( in case regex match another service )
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # make sure the service backend should NOT be there any longer
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        # TODO : test that coming back actually works

        self.service_if_pool.expose_services([])
        # every withhold service should NOT be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # service backend has not been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        services, service_types = self.get_system_state()
        dt = self.service_if_pool.update(services, service_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every withhold service should NOT be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # make sure the service backend has been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

    def test_service_update_disappear_withhold(self):
        """
        Test service exposing functionality for a service which already exists in
        the ros environment. Simple Normal usecase
        Sequence : UPDATE -> DISAPPEAR -> WITHHOLD
        :return:
        """

        servicename = '/test/absentsrv1'
        # every added service should be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # the backend should not have been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        self.service_if_pool.expose_services([servicename])
        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # service backend has not been created
        self.assertTrue(servicename not in self.service_if_pool.services.keys())

        dt = DiffTuple([], [])
        # create the service and then try updating again, simulating
        # it coming online after expose call.
        nonexistent_srv = rospy.Service(servicename, EmptySrv, srv_cb)
        try:
            # wait here until service actually appear in cache proxy
            with Timeout(5) as t:
                while not t.timed_out and nonexistent_srv.resolved_name not in dt.added:
                    services, service_types = self.get_system_state()
                    dt = self.service_if_pool.update(services, service_types)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(nonexistent_srv.resolved_name in dt.added)  # nonexistent_srv added
            self.assertEqual(dt.removed, [])  # nothing removed

            # every added service should be in the list of args
            self.assertTrue(servicename in self.service_if_pool.services_args)
            # service backend has been created
            self.assertTrue(servicename in self.service_if_pool.services.keys())

        # up to here possible sequences should have been already tested by previous tests
        # Now comes our actual disappearance / withholding test
        finally:
            nonexistent_srv.shutdown('testing disappearing service')

        # every added service should be in the list of args
        self.assertTrue(servicename in self.service_if_pool.services_args)
        # the backend should STILL be there
        self.assertTrue(servicename in self.service_if_pool.services.keys())
        # Note the service implementation should take care of possible errors in this case

        self.service_if_pool.expose_services([])
        # every withhold service should NOT be in the list of args
        self.assertTrue(servicename not in self.service_if_pool.services_args)
        # service backend should NOT be there any longer
        self.assertTrue(servicename not in self.service_if_pool.services.keys())


#TODO : here we always test the full update => test the diff algorithm as well !

@nose.tools.istest
class TestRosInterface1NoCache(TestRosServiceIfPool):

    @nose.tools.nottest
    def get_system_state(self):
        publishers, subscribers, services = self._master.getSystemState()[2]
        service_types = []  # master misses this API to be consistent
        return services, service_types

    def setUp(self):
        self.strpub = rospy.Publisher('/test/string', String, queue_size=1)
        self.emppub = rospy.Publisher('/test/empty', Empty, queue_size=1)

        self._master = rospy.get_master()
        self.service_if_pool = RosServiceIfPool()

    def tearDown(self):
        self.service_if_pool = None


# Testing with Connection Cache
@nose.tools.istest
class TestRosInterfaceCache(TestRosServiceIfPool):

    @nose.tools.nottest
    def get_system_state(self):
        # Note getting the system state via this interface, there is no need for callback
        publishers, subscribers, services = self.connection_cache.getSystemState()
        service_types = self.connection_cache.getServiceTypes()
        # handling fallback here since master doesnt have the API

        return services, service_types

    def setUp(self):

        # first we setup our publishers and our node (used by rospy.resolve_name calls to remap topics)
        self.strpub = rospy.Publisher('/test/string', String, queue_size=1)
        self.emppub = rospy.Publisher('/test/empty', Empty, queue_size=1)

        self._master = rospy.get_master()
        self.service_if_pool = RosServiceIfPool()

        # we need to speed fast enough for the tests to not fail on timeout...
        rospy.set_param('/connection_cache/spin_freq', 2)  # 2 Hz
        self.connection_cache_node = roslaunch.core.Node('rocon_python_comms', 'connection_cache.py',
                                                         name='connection_cache',
                                                         remap_args=[('~list', rospy.resolve_name('~connections_list')),
                                                                     (
                                                                     '~diff', rospy.resolve_name('~connections_diff'))])

        # Easier to remap the node topic to the proxy ones, instead of the opposite, since there is no dynamic remapping.
        # However for normal usecase, remapping the proxy handles is preferable.
        try:
            self.connection_cache_proc = self.launch.launch(self.connection_cache_node)
        except roslaunch.RLException as rlexc:
            raise nose.SkipTest("Connection Cache Node not found (part of rocon_python_comms pkg). Skipping test.")

        assert self.connection_cache_proc.is_alive()

        # wait for node to be started
        node_api = None
        with Timeout(5) as t:
            while not t.timed_out and node_api is None:
                node_api = rosnode.get_api_uri(rospy.get_master(), 'connection_cache')

        assert node_api is not None  # make sure the connection cache node is started before moving on.

        self.connection_cache = connection_cache_proxy_create()

    def tearDown(self):
        self.service_if_pool = None

        self.connection_cache_proc.stop()
        while self.connection_cache_proc.is_alive():
            time.sleep(0.2)  # waiting for cache node to die
        assert not self.connection_cache_proc.is_alive()
        time.sleep(1)  # TODO : investigate : we shouldnt need this


if __name__ == '__main__':

    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('test_service_if_pool_no_cache', 'test_all', TestRosInterface1NoCache)

    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('test_service_if_pool_cache', 'test_all', TestRosInterfaceCache)




#!/usr/bin/env python
from __future__ import absolute_import

import os
import sys
import logging


# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
# if not current_path in sys.path:
sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# Unit test import (  will emulate ROS setup if needed )
import time
from pyros import PyrosROS

# we import pyros.rosinterface here only to get the ROs setup done if needed,
# because we need to use ros modules for testing here
import pyros.rosinterface

import rospy
import roslaunch
import rosnode
from std_msgs.msg import String, Empty
from std_srvs.srv import Empty as EmptySrv, Trigger

import pyzmp

from pyros.rosinterface.rostests import Timeout

# useful test tools
from pyros_utils import rostest_nose
import unittest
import nose

# Ref : http://pythontesting.net/framework/nose/nose-fixture-reference/


# Fixtures should have the same effect as the <name>.test file for rostest.
# Should be used only by nose ( or other python test tool )
# CAREFUL with comments, copy paste mistake are real...
# CAREFUL dont use assertFalse -> easy to miss when reviewing code
def setup_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_setup_module()

        # Note : rospy.init_node is forbidden here because it would prevent the PyrosROS Process to start()


def teardown_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_teardown_module()


# Very basic echo service implementation for tests
def srv_cb(req):
    return req.request


@nose.tools.nottest
class TestPyrosROS(object):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """
    launch = None

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        # Start roslaunch
        TestPyrosROS.launch = roslaunch.scriptapi.ROSLaunch()
        TestPyrosROS.launch.start()

        # Note : rospy.init_node is forbidden here because it would prevent the PyrosROS Process to start()
        # Also we cannot use rospy.init_node in more than 1 test here

    @classmethod
    def teardown_class(cls):
        pass

    # Methods fixtures ( once per test method )
    def setUp(self, enable_cache=False):
        self.strpub = rospy.Publisher('/test/string', String, queue_size=1)
        self.emppub = rospy.Publisher('/test/empty', Empty, queue_size=1)

        self.enable_cache = enable_cache

    def tearDown(self):
        pass

    # @nose.SkipTest  # to help debugging ( FIXME : how to programmatically start only one test - maybe in fixture - ? )
    @nose.tools.timed(5)
    def test_rosnode_creation_termination(self):
        rosn = PyrosROS()
        nose.tools.assert_true(not rosn.is_alive())
        try:
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())

    @nose.tools.timed(5)
    def test_rosnode_provide_services(self):  # Here we check that this node actually provides all the services
        rosn = PyrosROS()
        nose.tools.assert_true(not rosn.is_alive())
        try:
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

            print("Discovering msg_build Service...")
            msg_build = pyzmp.discover("msg_build", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(msg_build is not None)
            print("msg_build providers : {svc}".format(svc=msg_build.providers))
            nose.tools.assert_equal(len(msg_build.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in msg_build.providers])

            print("Discovering topic Service...")
            topic = pyzmp.discover("topic", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(topic is not None)
            print("topic providers : {svc}".format(svc=topic.providers))
            nose.tools.assert_equal(len(topic.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in topic.providers])

            print("Discovering topics Service...")
            topics = pyzmp.discover("topics", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(topics is not None)
            print("topics providers : {svc}".format(svc=topics.providers))
            nose.tools.assert_equal(len(topics.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in topics.providers])

            print("Discovering service Service...")
            service = pyzmp.discover("service", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(service is not None)
            print("service providers : {svc}".format(svc=service.providers))
            nose.tools.assert_equal(len(service.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in service.providers])

            print("Discovering services Service...")
            services = pyzmp.discover("services", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(services is not None)
            print("services providers : {svc}".format(svc=services.providers))
            nose.tools.assert_equal(len(services.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in services.providers])

            print("Discovering param Service...")
            param = pyzmp.discover("param", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(param is not None)
            print("param providers : {svc}".format(svc=param.providers))
            nose.tools.assert_equal(len(param.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in param.providers])

            print("Discovering params Service...")
            params = pyzmp.discover("params", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(params is not None)
            print("params providers : {svc}".format(svc=params.providers))
            nose.tools.assert_equal(len(params.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in params.providers])

            print("Discovering setup Service...")
            setup = pyzmp.discover("setup", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(setup is not None)
            print("setup providers : {svc}".format(svc=setup.providers))
            nose.tools.assert_equal(len(setup.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in setup.providers])
        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())

    # For BWCompat only
    def test_rosnode_topics(self):  # Here we check that this node actually discovers topics

        # Starting underlying system before
        rospy.set_param('/string_pub/topic_name', '~test_str_topic')  # private topic name to not mess things up too much
        rospy.set_param('/string_pub/test_message', 'testing topic discovery')
        string_pub_node = roslaunch.core.Node('pyros_test', 'string_pub_node.py', name='string_pub')
        string_pub_process = self.launch.launch(string_pub_node)
        try:
            # Starting PyrosROS with preconfigured topics,
            rosn = PyrosROS(kwargs={'topics': ['/string_pub/test_str_topic'], 'enable_cache': self.enable_cache})  # careful assuming the topic fullname here
            try:
                nose.tools.assert_true(not rosn.is_alive())
                rosn.start()
                nose.tools.assert_true(rosn.is_alive())

                print("Discovering topics Service...")
                topics = pyzmp.discover("topics", 5)  # we wait a bit to let it time to start
                nose.tools.assert_true(topics is not None)
                print("topics providers : {svc}".format(svc=topics.providers))
                nose.tools.assert_equal(len(topics.providers), 1)
                nose.tools.assert_true(rosn.name in [p[0] for p in topics.providers])

                res = topics.call()
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                with Timeout(15) as t:
                    while not t.timed_out and not '/string_pub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = topics.call()

                nose.tools.assert_true('/string_pub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed
            finally:
                # finishing PyrosROS process
                if rosn is not None and rosn.is_alive():
                    rosn.shutdown()

            nose.tools.assert_true(not rosn.is_alive())

        finally:
            # finishing string_pub_process
            if string_pub_process is not None and string_pub_process.is_alive():
                string_pub_process.stop()

            nose.tools.assert_true(not string_pub_process.is_alive())

    def test_rosnode_topics_setup(self):  # Here we check that this node actually provides all the services
        rosn = PyrosROS()
        try:
            nose.tools.assert_true(not rosn.is_alive())
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

            print("Discovering topics Service...")
            topics = pyzmp.discover("topics", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(topics is not None)
            print("topics providers : {svc}".format(svc=topics.providers))
            nose.tools.assert_equal(len(topics.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in topics.providers])

            res = topics.call()
            nose.tools.assert_true('test_topic' not in res.keys())  # test_topic has not been created, detected or exposed

            print("Discovering setup Service...")
            setup = pyzmp.discover("setup", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(setup is not None)
            print("setup providers : {svc}".format(svc=setup.providers))
            nose.tools.assert_equal(len(setup.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in setup.providers])

            rospy.set_param('/string_pub/topic_name', '~test_str_topic')
            rospy.set_param('/string_pub/test_message', 'testing topic discovery')
            string_pub_node = roslaunch.core.Node('pyros_test', 'string_pub_node.py', name='string_pub')
            string_pub_process = self.launch.launch(string_pub_node)
            try:

                new_config = setup.call(kwargs={
                    'services': [],
                    'topics': ['/string_pub/test_str_topic'],
                    'params': [],
                    'enable_cache': self.enable_cache
                })
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                res = topics.call()
                with Timeout(15) as t:
                    while not t.timed_out and not '/string_pub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = topics.call()

                nose.tools.assert_true('/string_pub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed

            finally:
                # finishing all processes
                if string_pub_process is not None and string_pub_process.is_alive():
                    string_pub_process.stop()

            nose.tools.assert_true(not string_pub_process.is_alive())
        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())


    def test_rosnode_publishers(self):  # Here we check that this node actually discovers topics

        # Starting underlying system before
        rospy.set_param('/string_pub/topic_name', '~test_str_topic')  # private topic name to not mess things up too much
        rospy.set_param('/string_pub/test_message', 'testing topic discovery')
        string_pub_node = roslaunch.core.Node('pyros_test', 'string_pub_node.py', name='string_pub')
        string_pub_process = self.launch.launch(string_pub_node)
        try:
            # Starting PyrosROS with preconfigured topics,
            rosn = PyrosROS(kwargs={'publishers': ['/string_pub/test_str_topic'], 'enable_cache': self.enable_cache})  # careful assuming the topic fullname here
            try:
                nose.tools.assert_true(not rosn.is_alive())
                rosn.start()
                nose.tools.assert_true(rosn.is_alive())

                print("Discovering publishers Service...")
                publishers = pyzmp.discover("publishers", 5)  # we wait a bit to let it time to start
                nose.tools.assert_true(publishers is not None)
                print("publishers providers : {svc}".format(svc=publishers.providers))
                nose.tools.assert_equal(len(publishers.providers), 1)
                nose.tools.assert_true(rosn.name in [p[0] for p in publishers.providers])

                res = publishers.call()
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                with Timeout(15) as t:
                    while not t.timed_out and not '/string_pub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = publishers.call()

                nose.tools.assert_true('/string_pub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed
            finally:
                # finishing PyrosROS process
                if rosn is not None and rosn.is_alive():
                    rosn.shutdown()

            nose.tools.assert_true(not rosn.is_alive())

        finally:
            # finishing string_pub_process
            if string_pub_process is not None and string_pub_process.is_alive():
                string_pub_process.stop()

            nose.tools.assert_true(not string_pub_process.is_alive())

    def test_rosnode_publishers_setup(self):  # Here we check that this node actually provides all the services
        rosn = PyrosROS()
        try:
            nose.tools.assert_true(not rosn.is_alive())
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

            print("Discovering publishers Service...")
            publishers = pyzmp.discover("publishers", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(publishers is not None)
            print("publishers providers : {svc}".format(svc=publishers.providers))
            nose.tools.assert_equal(len(publishers.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in publishers.providers])

            res = publishers.call()
            nose.tools.assert_true('test_topic' not in res.keys())  # test_topic has not been created, detected or exposed

            print("Discovering setup Service...")
            setup = pyzmp.discover("setup", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(setup is not None)
            print("setup providers : {svc}".format(svc=setup.providers))
            nose.tools.assert_equal(len(setup.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in setup.providers])

            rospy.set_param('/string_pub/topic_name', '~test_str_topic')
            rospy.set_param('/string_pub/test_message', 'testing topic discovery')
            string_pub_node = roslaunch.core.Node('pyros_test', 'string_pub_node.py', name='string_pub')
            string_pub_process = self.launch.launch(string_pub_node)
            try:

                new_config = setup.call(kwargs={
                    'services': [],
                    'publishers': ['/string_pub/test_str_topic'],
                    'params': [],
                    'enable_cache': self.enable_cache
                })
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                res = publishers.call()
                with Timeout(15) as t:
                    while not t.timed_out and not '/string_pub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = publishers.call()

                nose.tools.assert_true('/string_pub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed

            finally:
                # finishing all processes
                if string_pub_process is not None and string_pub_process.is_alive():
                    string_pub_process.stop()

            nose.tools.assert_true(not string_pub_process.is_alive())
        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())
        # TODO : do we need a test with subscriber ?

    def test_rosnode_subscribers(self):  # Here we check that this node actually discovers topics

        # Starting underlying system before
        rospy.set_param('/string_sub/topic_name', '~test_str_topic')  # private topic name to not mess things up too much
        rospy.set_param('/string_sub/test_message', 'testing topic discovery')
        string_sub_node = roslaunch.core.Node('pyros_test', 'string_sub_node.py', name='string_sub')
        string_sub_process = self.launch.launch(string_sub_node)
        try:
            # Starting PyrosROS with preconfigured subscribers,
            rosn = PyrosROS(kwargs={'subscribers': ['/string_sub/test_str_topic'], 'enable_cache': self.enable_cache})  # careful assuming the topic fullname here
            try:
                nose.tools.assert_true(not rosn.is_alive())
                rosn.start()
                nose.tools.assert_true(rosn.is_alive())

                print("Discovering subscribers Service...")
                subscribers = pyzmp.discover("subscribers", 5)  # we wait a bit to let it time to start
                nose.tools.assert_true(subscribers is not None)
                print("subscribers providers : {svc}".format(svc=subscribers.providers))
                nose.tools.assert_equal(len(subscribers.providers), 1)
                nose.tools.assert_true(rosn.name in [p[0] for p in subscribers.providers])

                res = subscribers.call()
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                with Timeout(15) as t:
                    while not t.timed_out and not '/string_sub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = subscribers.call()

                nose.tools.assert_true('/string_sub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed
            finally:
                # finishing PyrosROS process
                if rosn is not None and rosn.is_alive():
                    rosn.shutdown()

            nose.tools.assert_true(not rosn.is_alive())

        finally:
            # finishing string_pub_process
            if string_sub_process is not None and string_sub_process.is_alive():
                string_sub_process.stop()

            nose.tools.assert_true(not string_sub_process.is_alive())

    def test_rosnode_subscribers_setup(self):  # Here we check that this node actually provides all the services
        rosn = PyrosROS()
        try:
            nose.tools.assert_true(not rosn.is_alive())
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

            print("Discovering subscribers Service...")
            subscribers = pyzmp.discover("subscribers", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(subscribers is not None)
            print("subscribers providers : {svc}".format(svc=subscribers.providers))
            nose.tools.assert_equal(len(subscribers.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in subscribers.providers])

            res = subscribers.call()
            nose.tools.assert_true('test_topic' not in res.keys())  # test_topic has not been created, detected or exposed

            print("Discovering setup Service...")
            setup = pyzmp.discover("setup", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(setup is not None)
            print("setup providers : {svc}".format(svc=setup.providers))
            nose.tools.assert_equal(len(setup.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in setup.providers])

            rospy.set_param('/string_pub/topic_name', '~test_str_topic')
            rospy.set_param('/string_pub/test_message', 'testing topic discovery')
            string_sub_node = roslaunch.core.Node('pyros_test', 'string_sub_node.py', name='string_sub')
            string_sub_process = self.launch.launch(string_sub_node)
            try:

                new_config = setup.call(kwargs={
                    'services': [],
                    'subscribers': ['/string_sub/test_str_topic'],
                    'params': [],
                    'enable_cache': self.enable_cache
                })
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                res = subscribers.call()
                with Timeout(15) as t:
                    while not t.timed_out and not '/string_sub/test_str_topic' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = subscribers.call()

                nose.tools.assert_true('/string_sub/test_str_topic' in res.keys())  # test_topic has been created, detected and exposed

            finally:
                # finishing all processes
                if string_sub_process is not None and string_sub_process.is_alive():
                    string_sub_process.stop()

            nose.tools.assert_true(not string_sub_process.is_alive())
        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())

    def test_rosnode_services(self):  # Here we check that this node actually discovers services

        # Starting underlying system before
        rospy.set_param('/string_echo/topic_name', '~topic')  # private names to not mess things up too much
        rospy.set_param('/string_echo/echo_topic_name', '~echo_topic')
        rospy.set_param('/string_echo/echo_service_name', '~echo_service')

        string_echo_node = roslaunch.core.Node('pyros_test', 'echo.py', name='string_echo')
        try:
            string_echo_process = self.launch.launch(string_echo_node)
        except roslaunch.RLException as rlexc:
            logging.error("pyros_test is needed to run this test. Please verify that it is installed in your ROS environment")
            raise
        try:
            # Starting PyrosROS with preconfigured services to expose
            rosn = PyrosROS(kwargs={'services': ['/string_echo/echo_service'], 'enable_cache': self.enable_cache})  # careful assuming the service fullname here
            try:
                nose.tools.assert_true(not rosn.is_alive())
                rosn.start()
                nose.tools.assert_true(rosn.is_alive())

                print("Discovering services Service...")
                services = pyzmp.discover("services", 5)  # we wait a bit to let it time to start
                nose.tools.assert_true(services is not None)
                print("services providers : {svc}".format(svc=services.providers))
                nose.tools.assert_equal(len(services.providers), 1)
                nose.tools.assert_true(rosn.name in [p[0] for p in services.providers])

                res = services.call()
                # What we get here is non deterministic
                # however we can wait for service to be detected to make sure we get it after some time

                with Timeout(5) as t:
                    while not t.timed_out and not '/string_echo/echo_service' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = services.call()

                nose.tools.assert_true('/string_echo/echo_service' in res.keys())  # echo_service has been created, detected and exposed
            finally:
                # finishing PyrosROS process
                if rosn is not None and rosn.is_alive():
                    rosn.shutdown()

            nose.tools.assert_true(not rosn.is_alive())

        finally:
            # finishing string_pub_process
            if string_echo_process is not None and string_echo_process.is_alive():
                string_echo_process.stop()

        nose.tools.assert_true(not string_echo_process.is_alive())

    def test_rosnode_services_setup(self):  # Here we check that this node actually provides all the services
        rosn = PyrosROS()
        try:
            nose.tools.assert_true(not rosn.is_alive())
            rosn.start()
            nose.tools.assert_true(rosn.is_alive())

            print("Discovering services Service...")
            services = pyzmp.discover("services", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(services is not None)
            print("services providers : {svc}".format(svc=services.providers))
            nose.tools.assert_equal(len(services.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in services.providers])

            res = services.call()
            nose.tools.assert_true('echo_service' not in res.keys())  # test_topic has not been created, detected or exposed

            print("Discovering setup Service...")
            setup = pyzmp.discover("setup", 5)  # we wait a bit to let it time to start
            nose.tools.assert_true(setup is not None)
            print("setup providers : {svc}".format(svc=setup.providers))
            nose.tools.assert_equal(len(setup.providers), 1)
            nose.tools.assert_true(rosn.name in [p[0] for p in setup.providers])

            rospy.set_param('/string_echo/topic_name', '~topic')  # private names to not mess things up too much
            rospy.set_param('/string_echo/echo_topic_name', '~echo_topic')
            rospy.set_param('/string_echo/echo_service_name', '~echo_service')

            string_echo_node = roslaunch.core.Node('pyros_test', 'echo.py', name='string_echo')
            string_echo_process = self.launch.launch(string_echo_node)
            try:

                new_config = setup.call(kwargs={
                    'services': ['/string_echo/echo_service'],
                    'topics': [],
                    'params': [],
                    'enable_cache': self.enable_cache,
                })
                # What we get here is non deterministic
                # however we can wait for topic to be detected to make sure we get it after some time

                res = services.call()

                with Timeout(5) as t:
                    while not t.timed_out and not '/string_echo/echo_service' in res.keys():
                        rospy.rostime.wallsleep(1)
                        res = services.call()

                nose.tools.assert_true('/string_echo/echo_service' in res.keys())  # test_topic has been created, detected and exposed

            finally:
                # finishing all processes
                if string_echo_process is not None and string_echo_process.is_alive():
                    string_echo_process.stop()

            nose.tools.assert_true(not string_echo_process.is_alive())

        finally:
            # finishing PyrosROS process
            if rosn is not None and rosn.is_alive():
                rosn.shutdown()

        nose.tools.assert_true(not rosn.is_alive())


# TODO : test each service

# TODO : test the update() is actually throttled (careful about cache behavior : no throttling)

# TODO : Test appearing / disappearing ROS topics / services

@nose.tools.istest
class TestPyrosROS1NoCache(TestPyrosROS):

    # Methods fixtures ( once per test method )
    def setUp(self):
        super(TestPyrosROS1NoCache, self).setUp()

    def tearDown(self):
        super(TestPyrosROS1NoCache, self).tearDown()


# Testing with Connection Cache
@nose.tools.istest
class TestPyrosROSCache(TestPyrosROS):

    def setUp(self):
        # we need to speed fast enough for the tests to not fail on timeout...
        rospy.set_param('/connection_cache/spin_freq', 2)  # 2 Hz
        self.connection_cache_node = roslaunch.core.Node('rocon_python_comms', 'connection_cache.py', name='connection_cache',
                                                         remap_args=[('~list', '/pyros_ros/connections_list'),
                                                                     ('~diff', '/pyros_ros/connections_diff'),
                                                                     ])
        try:
            self.connection_cache_proc = self.launch.launch(self.connection_cache_node)
        except roslaunch.RLException as rlexc:
            raise nose.SkipTest("Connection Cache Node not found (part of rocon_python_comms pkg). Skipping test.")

        node_api = None
        with Timeout(5) as t:
            while not t.timed_out and node_api is None:
                node_api = rosnode.get_api_uri(rospy.get_master(), 'connection_cache')

        assert node_api is not None  # make sure the connection cache node is started before moving on.

        super(TestPyrosROSCache, self).setUp(enable_cache=True)

    def tearDown(self):
        super(TestPyrosROSCache, self).tearDown()

        self.connection_cache_proc.stop()
        while self.connection_cache_proc.is_alive():
            time.sleep(0.2)  # waiting for cache node to die
        assert not self.connection_cache_proc.is_alive()
        time.sleep(1)  # TODO : investigate : we shouldnt need this


if __name__ == '__main__':

    import nose
    nose.runmodule()







from __future__ import absolute_import

import os

import six
import time


import sys

from .basenode import PyrosBase
import logging
import unicodedata

from .utils import deprecated


from . import config


class PyrosROS(PyrosBase):

    """
    Interface with ROS.
    """
    #: Default configuration parameters.
    _default_config = {
        'TOPICS': [],
        'SERVICES': [],
        'PARAMS': [],
        'ROS_USE_CONNECTION_CACHE': False,
        'ROS_CONNECTION_CACHE_LIST_TOPIC': "/rocon/connection_cache/list",
        'ROS_CONNECTION_CACHE_DIFF_TOPIC': "/rocon/connection_cache/diff",
    }

    def __init__(self, name=None, argv=None, pyros_config=None, args=None, kwargs=None):
        """
        Initializes a ROS node inside a zmp.Node, providing all the easy multiprocess communication of ZMP for it
        :param name: the name of the process
        :param argv: the arguments passed for its creation (ROS args)
        :param pyros_config: the configuration to use to configure pyros
        :param args: arguments to pass to the interface_class initializer. If provided, the first one is used as the ROS node name.
        :param kwargs: keyword arguments to pass to the interface_class initializer. If provided the 'argv' key will override the other 'argv' parameter.
        """

        # removing name from argv to avoid overriding specified name unintentionally
        argv = [arg for arg in (argv or []) if not arg.startswith('__name:=')]
        # protecting rospy from unicode
        str_argv = [
            unicodedata.normalize('NFKD', arg).encode('ascii', 'ignore') if isinstance(arg, unicode) else str(arg) for
            arg in argv]

        # argv is special :
        # They can be set only once, since they are passed to rospy.node_init in the interface.
        # That behavior is implemented in setup()
        self.argv = str_argv

        args = args or ()
        kwargs = kwargs or {}

        super(PyrosROS, self).__init__(
            name=name or 'pyros_ros',  # Careful with name : we cannot restart a node with the same name again...
            interface_class=(__package__, '.rosinterface', 'RosInterface'),  # lazy class evaluation to delay imports
            args=args,  # we want to pass the name to the interface to init the node with that name
            kwargs=kwargs,
            instance_relative_config=True,
            default_config=self._default_config)  # we pass our specific default config

        # overriding default config with parameter provided
        self.config_handler.configure(config)  # configuring with our package default
        if pyros_config:
            self.config_handler.configure(pyros_config)  # configuring with argument passed form user





    # TODO: get rid of this to need one less client-node call
    # we need make the message type visible to client,
    # and have him convert to a python structure that we can then convert to a message
    # dynamically right when calling the service.
    @deprecated
    def msg_build(self, connec_name):
        msg = None
        if self.interface:
            if connec_name in self.interface.topics.keys():
                input_msg_type = self.interface.topics.get(connec_name).rostype
                msg = input_msg_type()
            elif connec_name in self.interface.services.keys():
                input_msg_type = self.interface.services.get(connec_name).rostype_req
                msg = input_msg_type()
        return msg

    # These should match the design of RostfulClient and Protocol so we are consistent between pipe and python API
    #BWCOMPAT
    def topic(self, name, msg_content=None):
        res = None
        if self.interface:
            if msg_content is not None and name in self.interface.subscribers.keys():
                self.interface.subscribers.get(name).publish(msg_content)
            elif name in self.interface.publishers.keys():
                res = self.interface.publishers.get(name).get(consume=False)
        return res

    def topics(self):
        topics_dict = {}
        if self.interface:
            # merging pubs and subs for BWCOMPAT (like in pyros_mock)
            topics = self.interface.publishers.copy()
            topics.update(self.interface.subscribers)

            for t, tinst in six.iteritems(topics):
                topics_dict[t] = tinst.asdict()
        return topics_dict

    def publisher(self, name):
        res = None
        if self.interface and name in self.interface.publishers.keys():
            res = self.interface.publishers.get(name).get(consume=False)
        return res

    def publishers(self):
        publishers_dict = {}
        if self.interface:
            for t, tinst in six.iteritems(self.interface.publishers):
                publishers_dict[t] = tinst.asdict()
        return publishers_dict

    def subscriber(self, name, msg_content):
        res = None
        if self.interface and name in self.interface.subscribers.keys():
            self.interface.subscribers.get(name).publish(msg_content)
        return res

    def subscribers(self):
        subscribers_dict = {}
        if self.interface:
            for t, tinst in six.iteritems(self.interface.subscribers):
                subscribers_dict[t] = tinst.asdict()
        return subscribers_dict

    def service(self, name, rqst_content=None):
        resp_content = None

        # FIXME : if the service is not exposed this returns None.
        # Cost a lot time to find the reason since client code doesnt check the answer.
        # Maybe returning error is better ?
        if self.interface and name in self.interface.services.keys():
            resp_content = self.interface.services.get(name).call(rqst_content)
        return resp_content

    ###

    def services(self):
        services_dict = {}
        if self.interface:
            for s, sinst in six.iteritems(self.interface.services):
                services_dict[s] = sinst.asdict()
        return services_dict

    def param(self, name, value=None):
        if self.interface and name in self.interface.params.keys():
            if value is not None:
                self.interface.params.get(name).set(value)
                value = None  # consuming the message
            else:
                value = self.interface.params.get(name).get()
        return value

    def params(self):
        params_dict = {}
        if self.interface:
            for p, pinst in six.iteritems(self.interface.params):
                params_dict[p] = pinst.asdict()
        return params_dict

    def setup(self, publishers=None, subscribers=None, services=None, topics=None, params=None, enable_cache=False):
        """
        Service to dynamically setup the node.
        Node we cannot pass the name here as it should be set only once, the first time
        """
        # we get self.name and self.argv from the duplicated parent process memory.
        # this will create self.interface
        super(PyrosROS, self).setup(node_name=self.name, publishers=publishers, subscribers=subscribers, services=services, topics=topics, params=params, enable_cache=enable_cache, argv=self.argv)

    def run(self, *args, **kwargs):
        """
        Running in a zmp.Node process, providing zmp.services
        """

        # TODO : install shutdown hook to shutdown if detected

        try:
            # This will instantiate the rosinterface, which will initialize the ros node.
            # this is all done here in the child process, to not pollute the original process context.
            # this spins with regular frequency
            super(PyrosROS, self).run(*args, **kwargs)  # we override parent run to add one argument to ros interface

        except KeyboardInterrupt:
            logging.warn('PyrosROS node stopped by keyboard interrupt')

    def shutdown(self, join=True, timeout=None):
        """
        Clean shutdown of the node.
        :param join: optionally wait for the process to end (default : True)
        :return: None
        """
        super(PyrosROS, self).shutdown(join, timeout=timeout)


PyrosBase.register(PyrosROS)



from __future__ import absolute_import

import abc

# TODO : move this out of interface, it should also be known by client since we transfer exceptions
# All Pyros Exception must be pickleable and have a message property
class PyrosException(Exception):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def message(self):
        return


#!/usr/bin/env python
from __future__ import absolute_import

import six
import sys
import logging
import os
import pickle

# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
# if not current_path in sys.path:
sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# Unit test import
from pyros.rosinterface import ServiceBack


# ROS imports should now work from ROS or from python (with or without ROS env setup - emulated if needed)
import rospy
import roslaunch
import rosservice

#useful test tools
from pyros_utils import rostest_nose
import inspect
import unittest
import nose
from nose.tools import assert_false, assert_true, assert_equal


# test node process not setup by default (rostest dont need it here)
pub_process = None
echo_process = None
slow_process = None


# This should have the same effect as the <name>.test file for rostest.
# Should be used only by nose ( or other python test tool )
def setup_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_setup_module()

        # Start roslaunch
        launch = roslaunch.scriptapi.ROSLaunch()
        launch.start()

        # start required nodes - needs to match the content of *.test files for rostest to match

        global pub_process, echo_process, slow_process

        rospy.set_param('/echo_node/echo_service_name', 'test_service')
        echo_node = roslaunch.core.Node('pyros_test', 'echo.py', name='echo_node')
        try:
            echo_process = launch.launch(echo_node)
        except roslaunch.RLException as rlexc:
            logging.error("pyros_test is needed to run this test. Please verify that it is installed in your ROS environment")
            raise
        rospy.set_param('/slow_node/slow_service_name', 'test_timeout_service')
        slow_node = roslaunch.core.Node('pyros_test', 'string_slow_node.py', name='slow_node')
        try:
            slow_process = launch.launch(slow_node)
        except roslaunch.RLException as rlexc:
            logging.error("pyros_test is needed to run this test. Please verify that it is installed in your ROS environment")
            raise
        # set required parameters - needs to match the content of *.test files for rostest to match
        rospy.set_param('/stringServiceTest/echo_service_name', 'test_service')
        rospy.set_param('/stringServiceTest/slow_service_name', 'test_timeout_service')

        # we still need a node to interact with topics
        rospy.init_node('TestStringService', anonymous=True, disable_signals=True)
        # CAREFUL : this should be done only once per PROCESS
        # Here we enforce TEST MODULE 1<->1 PROCESS. ROStest style


def teardown_module():
    if not rostest_nose.is_rostest_enabled():
        # finishing all process are finished
        if pub_process is not None:
            pub_process.stop()
        if echo_process is not None:
            echo_process.stop()
        if slow_process is not None:
            slow_process.stop()

        rostest_nose.rostest_nose_teardown_module()


class TestService(unittest.TestCase):
    """ Testing the TopicBack class with String message """
    # misc method
    def logPoint(self):
        currentTest = self.id().split('.')[-1]
        callingFunction = inspect.stack()[1][3]
        print('in {0!s} - {1!s}()'.format(currentTest, callingFunction))

    def service_wait_type(self, service_name, retries=5, retrysleep=1):
        resolved_service_name = rospy.resolve_name(service_name)
        service_type = rosservice.get_service_type(resolved_service_name)
        retry = 0
        while not service_type and retry < 5:
            print('Service {service} not found. Retrying...'.format(service=resolved_service_name))
            rospy.rostime.wallsleep(retrysleep)
            retry += 1
            service_type = rosservice.get_service_type(resolved_service_name)
        if retry >= retries:
            self.fail("service {0} not found ! Failing.".format(resolved_service_name))
        return service_type

    def setUp(self):
        self.logPoint()
        # Here we hook to the test service in supported ways
        self.echo_service_name = rospy.get_param("/stringServiceTest/echo_service_name", "")
        self.slow_service_name = rospy.get_param("/stringServiceTest/slow_service_name", "")
        # No need of parameter for that, any str should work
        self.test_message = "testing"

        try:
            # actual fixture stuff
            # looking for the service ( similar code than ros_interface.py )
            echo_service_type = self.service_wait_type(self.echo_service_name)

            # exposing the service for testing here
            self.echo_service = ServiceBack(self.echo_service_name, echo_service_type)
            # looking for the service ( similar code than ros_interface.py )
            slow_service_type = self.service_wait_type(self.slow_service_name)

            # exposing the service for testing here
            self.slow_service = ServiceBack(self.slow_service_name, slow_service_type)
        except KeyboardInterrupt:
            self.fail("Test Interrupted !")

    def tearDown(self):
        self.logPoint()
        pass

    def test_service(self):
        try:
            self.logPoint()

            print("calling : {msg} on service {service}".format(msg=self.test_message, service=self.echo_service.name))
            resp = self.echo_service.call({'request': self.test_message})
            # Assert there is no difference
            assert_true(len(set(six.iteritems(resp)) ^ set(six.iteritems({'response': self.test_message}))) == 0)

        except KeyboardInterrupt:
            self.fail("Test Interrupted !")

    #TODO
    # def test_slow_service_timeout(self):
    #
    #     try:
    #         self.logPoint()
    #
    #         print("calling : {msg} on service {service}".format(msg=self.test_message, service=self.slow_service.name))
    #         with self.assertRaises(TypeError):
    #         resp = self.slow_service.call(self.slow_service.rostype_req(self.test_message))
    #         self.assertIn(resp.response, [self.test_message])
    #
    #     except KeyboardInterrupt:
    #         self.fail("Test Interrupted !")


if __name__ == '__main__':
    print("ARGV : %r", sys.argv)
    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('pyros', 'testService', TestService, sys.argv)



from __future__ import absolute_import

import logging
import mock
from collections import namedtuple
from contextlib import contextmanager

from . import pyros_client, config
from .pyros_mock import PyrosMock
from .pyros_ros import PyrosROS


# A context manager to handle server process launch and shutdown properly.
# It also creates a communication channel and passes it to a client.
@contextmanager
def pyros_ctx(name='pyros',
              argv=None,  # TODO : think about passing ros arguments http://wiki.ros.org/Remapping%20Arguments
              mock_client=False,
              mock_node=False,
              pyros_config=None):

    pyros_config = pyros_config or config  # using internal config if no other config passed

    subproc = None
    if not mock_client:
        if mock_node:
            logging.warn("Setting up pyros mock node...")
            subproc = PyrosMock(name)
        else:
            logging.warn("Setting up pyros ROS node...")
            subproc = PyrosROS(name, argv).configure(pyros_config)

        client_conn = subproc.start()

    ctx = namedtuple("pyros_context", "client")
    if mock_client:

        logging.warn("Setting up pyros mock client...")
        with mock.patch(pyros_client.__name__ + '.PyrosClient', autospec=True) as client:
            yield ctx(client=client)
    else:

        logging.warn("Setting up pyros actual client...")
        yield ctx(client=pyros_client.PyrosClient(client_conn))

    if subproc is not None:
        subproc.shutdown()



from __future__ import absolute_import
from collections import namedtuple, deque

from ...baseinterface import TransientIf


class MockSubscriber(TransientIf):
    """
    MockSubscriber is a mock of the class handling conversion from Python API to subscriber call
    """

    def __init__(self, topic_name, topic_type, queue_size=1):

        # getting the fullname to make sure we start with /
        topic_name = topic_name if topic_name.startswith('/') else '/' + topic_name

        topic_type = topic_type.msgtype

        super(MockSubscriber, self).__init__(topic_name, topic_type)

        self.msg = deque([], queue_size)

        self.empty_cb = None

    def cleanup(self):
        pass

    def get(self, num=0, consume=False):
        if not self.msg:
            return None

        res = None
        #TODO : implement returning multiple messages
        if consume:
            res = self.msg.popleft()
            if 0 == len(self.msg) and self.empty_cb:
                self.empty_cb()
                #TODO : CHECK that we can survive here even if we get dropped from the topic list
        else:
            res = self.msg[0]

        return res

    #returns the number of unread message
    def unread(self):
        return len(self.msg)

    def set_empty_callback(self, cb):
        self.empty_cb = cb

    # potentially called by a different thread : keep it miminal, just queue things.
    def topic_callback(self, msg):
        self.msg.appendleft(msg)




from __future__ import absolute_import

import logging
import os
import unicodedata

import sys

import six

"""
Client to pyros node, Python style.
Required for multiprocess communication.
"""

# When importing this your environment should already be setup
# and pyzmp should be found (from ROS packages or from python packages)
import pyzmp


from .exceptions import PyrosException

# TODO : Requirement : Check TOTAL send/receive SYMMETRY.
# If needed get rid of **kwargs arguments in call. Makes the interface less obvious and can trap unaware devs.


class PyrosServiceNotFound(PyrosException):
    def __init__(self, message):
        super(PyrosServiceNotFound, self).__init__(message)
        self.excmsg = message

    @property
    def message(self):
        return self.excmsg


# CAREFUL : exceptions must be pickleable ( we need to pass all arguments to the superclass )
class PyrosServiceTimeout(PyrosException):
    def __init__(self, message):
        super(PyrosServiceTimeout, self).__init__(message)
        self.excmsg = message

    @property
    def message(self):
        return self.excmsg

PyrosException.register(PyrosServiceTimeout)


# TODO : provide a test client ( similar to what werkzeug/flask does )
# The goal is to make it easy for users of pyros to test and validate their library only against the client,
# without having to have all the ROS environment installed and setup, and running extra processing
# just for unit testing...
class PyrosClient(object):
    # TODO : improve ZMP to return the socket_bind address to point to the exact IPC/socket channel.
    # And pass it here, instead of assuming node name is unique...
    def __init__(self, node_name=None):
        # Link to only one Server
        self.node_name = node_name

        # Discover all Services. Wait for at least one, and make sure it s provided by our expected Server
        self.msg_build_svc = pyzmp.Service.discover('msg_build', 5)
        if self.msg_build_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.msg_build_svc.providers]
        ):
            raise PyrosServiceNotFound('msg_build')

        self.topic_svc = pyzmp.Service.discover('topic', 5)
        if self.topic_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.topic_svc.providers]
        ):
            raise PyrosServiceNotFound('topic')

        self.service_svc = pyzmp.Service.discover('service', 5)
        if self.service_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.service_svc.providers]
        ):
            raise PyrosServiceNotFound('service')

        self.param_svc = pyzmp.Service.discover('param', 5)
        if self.param_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.param_svc.providers]
        ):
            raise PyrosServiceNotFound('param')

        self.topics_svc = pyzmp.Service.discover('topics', 5)
        if self.topics_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.topics_svc.providers]
        ):
            raise PyrosServiceNotFound('topics')

        self.services_svc = pyzmp.Service.discover('services', 5)
        if self.services_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.services_svc.providers]
        ):
            raise PyrosServiceNotFound('services')

        self.params_svc = pyzmp.Service.discover('params', 5)
        if self.params_svc is None or (
            self.node_name is not None and
            self.node_name not in [p[0] for p in self.params_svc.providers]
        ):
            raise PyrosServiceNotFound('params')

    def buildMsg(self, connection_name, suffix=None):
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(connection_name, unicode):
            connection_name = unicodedata.normalize('NFKD', connection_name).encode('ascii', 'ignore')
        res = self.msg_build_svc.call(args=(connection_name,))
        return res

    def topic_inject(self, topic_name, _msg_content=None, **kwargs):
        """
        Injecting message into topic. if _msg_content, we inject it directly. if not, we use all extra kwargs
        :param topic_name: name of the topic
        :param _msg_content: optional message content
        :param kwargs: each extra kwarg will be put int he message is structure matches
        :return:
        """
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(topic_name, unicode):
            topic_name = unicodedata.normalize('NFKD', topic_name).encode('ascii', 'ignore')

        if _msg_content is not None:
            # logging.warn("injecting {msg} into {topic}".format(msg=_msg_content, topic=topic_name))
            res = self.topic_svc.call(args=(topic_name, _msg_content,))
        else:  # default kwargs is {}
            # logging.warn("injecting {msg} into {topic}".format(msg=kwargs, topic=topic_name))
            res = self.topic_svc.call(args=(topic_name, kwargs,))

        return res is None  # check if message has been consumed

    def topic_extract(self, topic_name):
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(topic_name, unicode):
            topic_name = unicodedata.normalize('NFKD', topic_name).encode('ascii', 'ignore')

        try:
            res = self.topic_svc.call(args=(topic_name, None,))
        except pyzmp.service.ServiceCallTimeout as exc:
            six.reraise(PyrosServiceTimeout("Pyros Service call timed out."), None, sys.exc_info()[2])

        # TODO : if topic_name not exposed, we get None as res.
        # We should improve that behavior (display warning ? allow auto -dynamic- expose ?)

        return res

    def service_call(self, service_name, _msg_content=None, **kwargs):
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(service_name, unicode):
            service_name = unicodedata.normalize('NFKD', service_name).encode('ascii', 'ignore')

        try:
            if _msg_content is not None:
                res = self.service_svc.call(args=(service_name, _msg_content,))
            else:  # default kwargs is {}
                res = self.service_svc.call(args=(service_name, kwargs,))
        except pyzmp.service.ServiceCallTimeout as exc:
            six.reraise(PyrosServiceTimeout("Pyros Service call timed out."), None, sys.exc_info()[2])
        # A service that doesn't exist on the node will return res_content.resp_content None.
        # It should probably except...
        # TODO : improve error handling, maybe by checking the type of res ?

        return res

    def param_set(self, param_name, _value=None, **kwargs):
        """
        Setting parameter. if _value, we inject it directly. if not, we use all extra kwargs
        :param topic_name: name of the topic
        :param _value: optional value
        :param kwargs: each extra kwarg will be put in the value if structure matches
        :return:
        """
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(param_name, unicode):
            param_name = unicodedata.normalize('NFKD', param_name).encode('ascii', 'ignore')

        _value = _value or {}

        if kwargs:
            res = self.param_svc.call(args=(param_name, kwargs,))
        elif _value is not None:
            res = self.param_svc.call(args=(param_name, _value,))
        else:   # if _msg_content is None the request is invalid.
                # just return something to mean False.
            res = 'WRONG SET'

        return res is None  # check if message has been consumed

    def param_get(self, param_name):
        #changing unicode to string ( testing stability of multiprocess debugging )
        if isinstance(param_name, unicode):
            param_name = unicodedata.normalize('NFKD', param_name).encode('ascii', 'ignore')
        res = self.param_svc.call(args=(param_name, None,))
        return res

    def topics(self):
        try:
            res = self.topics_svc.call(send_timeout=5000, recv_timeout=10000)  # Need to be generous on timeout in case we are starting up multiprocesses
        except pyzmp.service.ServiceCallTimeout as exc:
            six.reraise(PyrosServiceTimeout("Pyros Service call timed out."), None, sys.exc_info()[2])
        return res
        
    def services(self):
        try:
            res = self.services_svc.call(send_timeout=5000, recv_timeout=10000)  # Need to be generous on timeout in case we are starting up multiprocesses
        except pyzmp.service.ServiceCallTimeout as exc:
            six.reraise(PyrosServiceTimeout("Pyros Service call timed out."), None, sys.exc_info()[2])
        return res

    def params(self):
        res = self.params_svc.call(send_timeout=5000, recv_timeout=10000)  # Need to be generous on timeout in case we are starting up multiprocesses
        return res

    #def listacts(self):
    #    return {}

    # def namespaces(self):
    #    res = self.namespaces_svc.call(args=({},))
    #    return res

    # def interactions(self):
    #    res = self.interactions_svc.call(args=({},))
    #    return res

    # def interaction(self, name):
    #    res = self.interaction_svc.call(args=("",))
    #    return res

    # def has_rocon(self):
    #    res = self.has_rocon_svc.call(args=(False,))
    #    return res

# TODO : test client with Rostful Node ( and detect ROS or not to confirm behvior )



from __future__ import absolute_import
from __future__ import print_function


class Transient(object):
    """
    Transient is a base class handling transient description.
    This transient can be a service, a param, a publisher, etc.
    This class aims at making this transient usable from python, in a common way for all implementations.
    It will be manipulated from TransientPool, to track when it goes up and down and when it should be interfaced.
    """

    def __init__(self, transient_name, transient_type=None, transient_endpoints=None):
        """
        Building a transient from external data
        :param transient_name:
        :param transient_type:
        :param transient_endpoints: a set containing multiple endpoints tuples (node_name, uri, ...)
        Note the endpoints can be specific to the implementation
        """

        self.name = transient_name
        self.type = transient_type
        self.endpoints = transient_endpoints or set()




# CORE ROS API
# This is useful to quickly patch ROS problems
#
# Note : pyros_setup shouldn't be used here, since we need it for other packages installed in ROS environment
# and not only for ROS interface
#
# TODO : move this to pyros_utils after enough time has passed to make sure it s solid enough...



from __future__ import absolute_import

import sys
import os
import time

# # This is needed if running this test directly (without using nose loader)
# # prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# # And we need our current module to be found first.
# current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
# # if not current_path in sys.path:
# sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec


from pyros.pyros_ctx_server import pyros_ctx
from pyros.pyros_client import PyrosClient


def testPyrosMockCtx():
    with pyros_ctx(mock_node=True) as ctx:
        assert isinstance(ctx.client, PyrosClient)

    # TODO : assert the context manager does his job ( HOW ? )


def testPyrosROSCtx():
    # start ros system , before PyrosROS process and Client otherwise Client assumes there is  problem ( discovery timeout )
    # we might need to load pyros_setup here...
    try:
        import pyros_utils
    except ImportError:
        # TODO : find a proper ay to log from a test like here...
        #_logger.warning("loading pyros_setup and configuring your ROS environment")
        import pyros_setup
        # This will load the pyros_setup configuration from the environment
        pyros_setup.configurable_import().configure().activate()
        import pyros_utils

    master, roscore_proc = pyros_utils.get_master(spawn=True)  # we start the master if needed

    assert master.is_online()

    with pyros_ctx(mock_node=False) as ctx:

        assert isinstance(ctx.client, PyrosClient)

    # TODO : assert the context manager does his job ( HOW ? )

    if roscore_proc is not None:
        roscore_proc.terminate()
        while roscore_proc.is_alive():
            time.sleep(0.2)  # waiting for roscore to die


if __name__ == '__main__':

    import nose
    nose.runmodule()



from __future__ import absolute_import

import logging
from .api import rospy_safe as rospy

# create logger
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

try:
    import rocon_python_comms
except ImportError:
    rocon_python_comms = None


def connection_cache_proxy_create(_proxy_cb=None):
    """
    Creates and return a connection cache proxy if possible, handling all likely exceptions
    :param _proxy_cb:
    :return:
    """
    if rocon_python_comms is None:
        return None
    else:
        # connectioncache proxy if available (remap the topics if necessary instead of passing params)
        try:
            connection_cache_proxy = rocon_python_comms.ConnectionCacheProxy(
                list_sub='~connections_list',
                handle_actions=False,
                user_callback=_proxy_cb,
                diff_opt=True,
                diff_sub='~connections_diff'
            )

        except AttributeError as attr_exc:
            # attribute error (likely rocon_python_comms doesnt have ConnectionCacheProxy)
            # NOT EXPECTED System configuration problem : BE LOUD !
            # timeout initializing : disabling the feature but we should be LOUD about it
            rospy.logwarn("connection_cache_utils : FAILED during initialization of Connection Cache Proxy. Disabling.")
            import traceback
            rospy.logwarn('Exception: {0}'.format(traceback.format_stack()))
            return None

        except rocon_python_comms.ConnectionCacheProxy.InitializationTimeout as timeout_exc:

            # timeout initializing : disabling the feature but we should WARN about it
            rospy.logwarn("connection_cache_utils : TIMEOUT during initialization of Connection Cache Proxy. Disabling.")
            return None

        else:
            rospy.loginfo("connection_cache_utils : Connection Cache Optimization enabled")
            return connection_cache_proxy


def connection_cache_marshall(cb_ss):
    """
    Marshall data comming from connection cache callback to adapt to the internal data format for pyros
    :param cb_ss: system state to convert
    :param old_cb_ss: Old system state in case we want to merge with the recent one
    :return:
    """

    publishers = {}
    subscribers = {}
    topic_types = []
    services = {}
    service_types = []
    params = {}  # TODO cb_ss.params

    # We assume we get SystemState type from the connection cache proxy (directly, the first time)

    for k, v in cb_ss.services.iteritems():
        services[k] = services.get(k, set()) | v.nodes

    for k, v in cb_ss.publishers.iteritems():
        publishers[k] = publishers.get(k, set()) | v.nodes

    for k, v in cb_ss.subscribers.iteritems():
        subscribers[k] = subscribers.get(k, set()) | v.nodes

    pubset = {(name, chan.type) for name, chan in cb_ss.publishers.iteritems()}
    subset = {(name, chan.type) for name, chan in cb_ss.subscribers.iteritems()}
    topic_types += [list(t) for t in (pubset | subset)]

    svcset = {(name, chan.type) for name, chan in cb_ss.services.iteritems()}
    service_types += [list(t) for t in svcset]

    return {
        'publishers': publishers,
        'subscribers': subscribers,
        'services': services,
        'params': params,
        'topic_types': topic_types,
        'services_types': service_types
    }


def connection_cache_merge_marshalled(marshalled_ss, old_marshalled_ss=None):

    old_marshalled_ss = old_marshalled_ss or {}

    publishers = old_marshalled_ss.get('publishers', {})
    subscribers = old_marshalled_ss.get('subscribers', {})
    topic_types = old_marshalled_ss.get('topic_types', [])
    services = old_marshalled_ss.get('services', {})
    service_types = old_marshalled_ss.get('service_types', [])
    params = {}  # TODO cb_ss.params

    # We come back again here after marshalling already once...
    for k, endpoints in marshalled_ss.get('services', {}).iteritems():
        services[k] = services.get(k, set()) | endpoints

    for k, endpoints in marshalled_ss.get('publishers', {}).iteritems():
        publishers[k] = publishers.get(k, set()) | endpoints

    for k, endpoints in marshalled_ss.get('subscribers', {}).iteritems():
        subscribers[k] = subscribers.get(k, set()) | endpoints

    topic_types += marshalled_ss.get('topic_types', [])
    service_types += marshalled_ss.get('service_types', [])

    return {
        'publishers': publishers,
        'subscribers': subscribers,
        'services': services,
        'params': params,
        'topic_types': topic_types,
        'services_types': service_types
    }


from __future__ import absolute_import

from collections import OrderedDict
from importlib import import_module

import roslib

from .api import rospy_safe as rospy
from .message_conversion import get_msg, get_msg_dict, populate_instance, extract_values, FieldTypeMismatchException, NonexistentFieldException
from ..baseinterface import TransientIf


# outputs message structure as string (useful ?)
def get_service_srv(service):
    return '\n'.join([
        get_msg(service.rostype_req),
        '---',
        get_msg(service.rostype_resp),
    ])


# outputs message structure as dict
def get_service_srv_dict(service):
    return get_msg_dict(service.rostype_req), get_msg_dict(service.rostype_resp)


class ServiceTuple(object):
    """
    Class representing the service the interface connects to
    """
    def __init__(self, name, type):
        self.name = name
        self.type = type
# Note : for service the connection endpoint is not important
# R1 : service concept is not connection oriented ( client is inexistant until service is used )
# R2 : ROS only keeps the last service provider in master, previous ones are just erased.
# TODO: make that the pickled representation of ServiceBack (check asdict())


class ServiceBack(TransientIf):
    """
    ServiceBack is the class handling conversion from Python API to ROS Service
    """
    def __init__(self, service_name, service_type):

        service_name = rospy.resolve_name(service_name)
        super(ServiceBack, self).__init__(service_name, service_type)

        service_type_module, service_type_name = tuple(service_type.split('/'))
        roslib.load_manifest(service_type_module)
        srv_module = import_module(service_type_module + '.srv')

        self.rostype_name = service_type
        self.rostype = getattr(srv_module, service_type_name)
        self.rostype_req = getattr(srv_module, service_type_name + 'Request')
        self.rostype_resp = getattr(srv_module, service_type_name + 'Response')

        self.srvtype = get_service_srv_dict(self)

        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Adding service interface {name} {typename}".format(
                name=self.name, typename=self.rostype_name))

        self.proxy = rospy.ServiceProxy(self.name, self.rostype)

    def cleanup(self):

        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Removing service interface {name} {typename}".format(
                name=self.name, typename=self.rostype))

        super(ServiceBack, self).cleanup()

    def asdict(self):
        """
        Here we provide a dictionary suitable for a representation of the Topic instance
        the main point here is to make it possible to transfer this to remote processes.
        We are not interested in pickleing the whole class with Subscriber and Publisher
        :return:
        """

        return OrderedDict({
            'name': self.name,
            'fullname': self.name,  # for BWcompat
            'rostype_name': self.rostype_name,
            'srvtype': self.srvtype,
        })

    def call(self, rosreq_content=None):
        try:
            rqst = self.rostype_req()
            populate_instance(rosreq_content, rqst)

            fields = []
            if rosreq_content:
                for slot in rqst.__slots__:
                    fields.append(getattr(rqst, slot))
                fields = tuple(fields)

            resp = self.proxy(*fields)
            resp_content = extract_values(resp)

            return resp_content

        except rospy.ServiceException as e:
            rospy.logerr("[{name}] : service exception {e}".format(name=__name__, e=e))
            raise

        except FieldTypeMismatchException as e:
            rospy.logerr("[{name}] : field type mismatch {e}".format(name=__name__, e=e))
            raise
        except NonexistentFieldException as e:
            rospy.logerr("[{name}] : non existent field {e}".format(name=__name__, e=e))
            raise





from __future__ import absolute_import, print_function

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..')))

import nose
from nose.tools import assert_true, assert_false, assert_raises
import unittest

from pyros.rosinterface.mock import extract_values, populate_instance, FieldTypeMismatchException, NonexistentFieldException, StatusMsg
from pyros.rosinterface.mock import MockSystem


class TestMockService(unittest.TestCase):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        cls.system = MockSystem()

    @classmethod
    def teardown_class(cls):
        pass

    def test_echo_fortytwo(self):
        param = self.system.create_parameter('random_param', int)
        assert_true(param.setval(42))
        recv = param.getval()
        print(recv)
        assert_true(recv == 42)


if __name__ == '__main__':

    nose.runmodule()



from __future__ import absolute_import

import logging
from itertools import ifilter

import rosservice

from .api import rospy_safe as rospy

# create logger
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

from ..baseinterface import TransientIfPool

from .service_if import ServiceBack, ServiceTuple

try:
    import rocon_python_comms
except ImportError:
    rocon_python_comms = None


class RosServiceIfPool(TransientIfPool):

    """
    MockInterface.
    """
    def __init__(self, services=None):
        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        super(RosServiceIfPool, self).__init__(services, transients_desc="services")

    def get_transients_available(self):  # function returning all services available on the system
        return self.available

    def transient_type_resolver(self, service_name):  # function resolving the type of a service
        svc = self.available.get(service_name)
        if svc:
            if svc.type is None:  # if the type is unknown, lets discover it (needed to interface)
                try:
                    resolved_service_name = rospy.resolve_name(service_name)  # required or not ?
                    svc.type = rosservice.get_service_type(resolved_service_name)
                except rosservice.ROSServiceIOException:  # exception can occur -> just reraise
                    raise
            return svc.type  # return the type
        else:
            rospy.logerr("ERROR while resolving {service_name}. Service not known as available. Ignoring".format(**locals()))
            return None

    def TransientMaker(self, service_name, service_type, *args, **kwargs):  # the service class implementation
        return ServiceBack(service_name, service_type, *args, **kwargs)

    def TransientCleaner(self, service):  # the service class cleanup implementation
        return service.cleanup()

    ## bwcompat
    # REQUESTED
    @property
    def services_args(self):
        return self.transients_args


    # AVAILABLE
    @property
    def services_available(self):
        return self.available

    # INTERFACED
    @property
    def services(self):
        return self.transients

    # EXPOSE
    def expose_services(self, svc_regex):
        return self.expose_transients_regex(svc_regex)

    def get_svc_available(self):  # function returning all services available on the system
        return self.get_transients_available

    def service_type_resolver(self, service_name):  # function resolving the type of a service
        return self.transient_type_resolver(service_name)


    def ServiceMaker(self, service_name, service_type):  # the service class implementation
        return self.TransientMaker(service_name, service_type)

    def ServiceCleaner(self, service):  # the service class cleanup implementation
        return self.TransientCleaner(service)
    ###########

    def reset_state(self, services, service_types):
        """
        Reset internal system state representation.
        expect lists in format similar to masterAPI.
        :param publishers:
        :param subscribers:
        :param services:
        :param topic_types:
        :param service_types:
        :return:
        """
        self.available = dict()
        for s in services:  # We assume s[1] is never empty here
            st = next(ifilter(lambda lst: s[0] == lst[0], service_types), [])
            stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
            self.available[stp.name] = stp

        # We still need to return DiffTuples
        return services

    def compute_state(self, services_dt, service_types_dt):
        """
        This is called only if there is a cache proxy with a callback, and expects DiffTuple filled up with names or types
        :param services_dt:
        :param publishers_dt:
        :param subscribers_dt:
        :return:
        """
        for s in services_dt.added:
            st = next(ifilter(lambda lst: s[0] == lst[0], service_types_dt.added), [])
            stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
            if stp.name in self.available:
                if self.available[stp.name].type is None and stp.type is not None:
                    self.available[stp.name].type = stp.type
            else:
                self.available[stp.name] = stp

        for s in services_dt.removed:
            st = next(ifilter(lambda lst: s[0] == lst[0], service_types_dt.removed), [])
            stp = ServiceTuple(name=s[0], type=st[1] if len(st) > 0 else None)
            if stp.name in self.available:
                self.available.pop(stp.name, None)

        # We still need to return DiffTuples
        return services_dt

    # for use with line_profiler or memory_profiler
    # Not working yet... need to solve multiprocess profiling issues...
    # @profile
    def update_delta(self, services_dt, service_types_dt=None):
        services_dt = self.compute_state(services_dt, service_types_dt or [])

        if services_dt.added or services_dt.removed:
            _logger.debug(
                rospy.get_name() + " Services Delta {services_dt}".format(**locals()))

        # update_services wants only names
        dt = self.transient_change_diff(
            transient_appeared=[s[0] for s in services_dt.added],
            transient_gone=[s[0] for s in services_dt.removed]
            # add_names=regexes_match_sublist(self.transients_args, [s[0] for s in services_dt.added]),
            # remove_names=[s[0] for s in services_dt.removed if s[0] not in self.get_transients_available()]
        )

        if dt.added or dt.removed:
            _logger.debug(
                rospy.get_name() + " Update Delta {dt}".format(**locals()))
        return dt


    # for use with line_profiler or memory_profiler
    # Not working yet... need to solve multiprocess profiling issues...
    # @profile
    def update(self, services, service_types):
        # First we need to reflect the external system state in internal cache
        self.reset_state(services, service_types)

        if services:
            _logger.debug(
                rospy.get_name() + " Services List {services}".format(**locals()))

        # Second we update our interfaces based on that new system state
        # TODO : pass full services state here to avoid having to retrieve indirectly
        dt = self.transient_change_detect()

        if dt.added or dt.removed:
            _logger.debug(
                rospy.get_name() + " Update Delta {dt}".format(**locals()))

        return dt


TransientIfPool.register(RosServiceIfPool)




#!/usr/bin/env python
from __future__ import absolute_import

# ROS SETUP if needed
import logging
import multiprocessing
import time
import cProfile

import sys
from pyros import PyrosROS
import rospy
import rosgraph
import roslaunch

roscore_process = None

# BROKEN ?? start roscore beofre running this...
# if not rosgraph.masterapi.is_online():
#     # Trying to solve this : http://answers.ros.org/question/215600/how-can-i-run-roscore-from-python/
#     def ros_core_launch():
#         roslaunch.main(['roscore', '--core'])  # same as rostest_main implementation
#
#     roscore_process = multiprocessing.Process(target=ros_core_launch)
#     roscore_process.start()
#
#     while not roscore_process.is_alive():
#         time.sleep(0.2)  # waiting for roscore to be born
#
#     assert roscore_process.is_alive()

assert rosgraph.masterapi.is_online()

# Start roslaunch
launch = roslaunch.scriptapi.ROSLaunch()
launch.start()

# starting connection cache is available
rospy.set_param('/connection_cache/spin_freq', 2)  # 2 Hz
connection_cache_node = roslaunch.core.Node('rocon_python_comms', 'connection_cache.py', name='connection_cache',
                                                 remap_args=[('~list', '/pyros_ros/connections_list'),
                                                             ('~diff', '/pyros_ros/connections_diff'),
                                                             ])
try:
    connection_cache_proc = launch.launch(connection_cache_node)
except roslaunch.RLException as rlexc:
    pass  # ignore

time.sleep(2)

# start a bunch of node (this will load ros interface)
pub_proc = []

def start_pub_node(pubnum):
    node_name = 'string_pub_node_' + str(pubnum)
    rospy.set_param('/' + node_name + '/topic_name', 'pub_' + str(pubnum))
    rospy.set_param('/' + node_name + '/test_message', 'msg_' + str(pubnum))
    node = roslaunch.core.Node('pyros_test', 'string_pub_node.py', name=node_name)
    try:
        pub_proc.append(launch.launch(node))
    except roslaunch.RLException as rlexc:
        logging.error(
            "pyros_test is needed to run this. Please verify that it is installed in your ROS environment")
        raise

# TODO : make MANY nodes / services / params to simulate complex robot and make profiling more realistic.
time.sleep(2)  # waiting for nodes to be up

rosn = PyrosROS()

rosn.setup(
    services=['/test/empsrv', '/test/trgsrv'],
    params=['/test/confirm_param'],
    enable_cache=connection_cache_proc.is_alive()
)

print("Module LOADED")


def update_loop():
    total_count = 1024*1024*255
    count = 0
    start = time.time()
    pct = 0
    last_pct = -1
    max_pubnodes = 42
    node_step = 0
    last_node_step = -1
    while count < total_count:
        # time is ticking
        now = time.time()
        timedelta = now - start
        start = now

        rosn.update(timedelta)

        count += 1

        # creating and removing nodes while looping
        node_step = count * max_pubnodes * 2/ total_count
        if node_step != last_node_step:
            last_node_step = node_step
            if count < total_count/2:
                # adding nodes
                print("adding node {0}".format(node_step))
                start_pub_node(node_step)

            elif pub_proc:
                # stopping nodes LIFO
                print("stopping node {0}".format(len(pub_proc)-1))
                pub_proc.pop().stop()

        pct = count * 100 / total_count
        if pct != last_pct:
            last_pct = pct
            sys.stdout.write("\r" + str(last_pct) + "%")
            sys.stdout.flush()

# In case you want to run kernprof here
#update_loop()
cProfile.run('update_loop()', sort='cumulative')

# ensuring all process are finished
for p in pub_proc:
    p.stop()

if connection_cache_proc is not None:
    connection_cache_proc.stop()

rospy.signal_shutdown('test complete')

if roscore_process is not None:
    roscore_process.terminate()  # make sure everything is stopped


#!/usr/bin/env python
from __future__ import absolute_import

import os
import sys
import pickle

# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
import time

# current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
# # if not current_path in sys.path:
# sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# setting up logging from this test
import logging.config
# Setting up logging for this test
# TODO : solve multiprocess problems... this seems not enough to get on console...
logging.config.dictConfig(
    {
        'version': 1,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(name)s:%(message)s'
            },
        },
        'handlers': {
            'null': {
                'level': 'DEBUG',
                'class': 'logging.NullHandler',
            },
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
        },
        'loggers': {
            'pyros_config': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
            'pyros_setup': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'pyros': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'pyros.rosinterface': {
                'handlers': ['console'],
                'level': 'DEBUG',
            },
            'pyros.ros_interface': {
                'handlers': ['console'],
                'level': 'DEBUG',
            }
        }
    }
)


# Unit test import (  will emulate ROS setup if needed )
import nose

from pyros.baseinterface import DiffTuple
from pyros.rosinterface import RosSubscriberIfPool
from pyros.rosinterface.connection_cache_utils import connection_cache_proxy_create, connection_cache_marshall


import rospy
import roslaunch
import rosnode
from std_msgs.msg import String, Empty
from std_srvs.srv import Empty as EmptySrv, Trigger

from pyros.rosinterface.rostests import Timeout

# useful test tools
from pyros_utils import rostest_nose
import unittest


# This should have the same effect as the <name>.test file for rostest.
# Should be used only by nose ( or other python test tool )
# CAREFUL with comments, copy paste mistake are real...
# CAREFUL dont use assertFalse -> easy to miss when reviewing code
def setup_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_setup_module()

    # Using pubs and subs require us to be a node
    rospy.init_node('test_topic_if_pool', argv=None, disable_signals=True)
    # CAREFUL : rospy.init_node should be done only once per PROCESS
    # Here we enforce TEST RUN 1<->1 MODULE 1<->1 PROCESS. ROStest style.


def teardown_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_teardown_module()


@nose.tools.nottest
class TestRosSubscriberIfPool(unittest.TestCase):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """

    launch = None
    # test node process not setup by default (rostest dont need it here)
    empty_srv_process = None
    trigger_srv_process = None

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        # Start roslaunch
        TestRosSubscriberIfPool.launch = roslaunch.scriptapi.ROSLaunch()
        TestRosSubscriberIfPool.launch.start()

        # start required nodes - needs to match the content of *.test files for rostest to match
        global empty_srv_process, trigger_srv_process
        empty_srv_node = roslaunch.core.Node('pyros_test', 'emptyService.py', name='empty_service')
        trigger_srv_node = roslaunch.core.Node('pyros_test', 'triggerService.py', name='trigger_service')
        TestRosSubscriberIfPool.empty_srv_process = TestRosSubscriberIfPool.launch.launch(empty_srv_node)
        TestRosSubscriberIfPool.trigger_srv_process = TestRosSubscriberIfPool.launch.launch(trigger_srv_node)

    @classmethod
    def teardown_class(cls):
        # ensuring all process are finished
        if TestRosSubscriberIfPool.empty_srv_process is not None:
            TestRosSubscriberIfPool.empty_srv_process.stop()
        if TestRosSubscriberIfPool.trigger_srv_process is not None:
            TestRosSubscriberIfPool.trigger_srv_process.stop()

    # EXPOSE SUBSCRIBERS + UPDATE Interface
    def test_subscriber_appear_expose_update(self):
        """
        Test topic exposing functionality for a topic which already exists in
        the ros environment. Simple Normal usecase
        Sequence : APPEAR -> EXPOSE -> UPDATE
        :return:
        """
        topicname = '/test/string'
        self.subscriber_if_pool.expose_subscribers([topicname])
        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # topic backend has not been created since the update didn't run yet
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        dt = DiffTuple([], [])
        # NOTE : We need to wait to make sure the tests nodes are started...
        with Timeout(5) as t:
            while not t.timed_out and topicname not in dt.added:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(topicname in dt.added)  # has been detected

        # every exposed topic should remain in the list of args ( in case regex match another topic )
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has been created
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        # cleaning up
        self.subscriber_if_pool.expose_subscribers([])
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend has not been created since the update didn't run yet
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

    def test_subscriber_appear_update_expose(self):
        """
        Test topic exposing functionality for a topic which already exists in the ros environment.
        Normal usecase
        Sequence : (UPDATE?) -> APPEAR -> UPDATE -> EXPOSE (-> UPDATE?)
        :return:
        """

        topicname = '/test/nonexistent1'
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        # First update should not change state
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # create the publisher and then try exposing the topic again, simulating
        # it coming online before expose call.
        nonexistent_pub = rospy.Subscriber(topicname, Empty, queue_size=1)
        with Timeout(5) as t:
            while not t.timed_out and nonexistent_pub.resolved_name not in self.subscriber_if_pool.subscribers_available:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.added, [])  # nothing added (not exposed yet)
                self.assertEqual(dt.removed, [])  # nothing removed
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        # TODO : do we need a test with subscriber ?

        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        dt = self.subscriber_if_pool.expose_subscribers([topicname])
        # every exposed topic should remain in the list of args ( in case regex match another topic )
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has been created
        self.assertTrue(topicname in dt.added)
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        # removing publisher
        nonexistent_pub.unregister()  # https://github.com/ros/ros_comm/issues/111 ( topic is still registered on master... )

        # and update should be enough to cleanup
        with Timeout(5) as t:
            while not t.timed_out and not topicname in dt.removed:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.added, [])  # nothing added
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(topicname in dt.removed)
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_available)

        # every exposed topic should still be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # the backend should not be there any longer
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

    def test_subscriber_expose_appear_update(self):
        """
        Test basic topic adding functionality for a topic which does not yet exist
        in the ros environment ( + corner cases )
        Sequence : (UPDATE? ->) -> EXPOSE -> (UPDATE? ->) APPEAR -> UPDATE
        :return:
        """
        topicname = '/test/nonexistent2'
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        # First update should not change state
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertTrue(topicname not in dt.added)  # not detected
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        self.subscriber_if_pool.expose_subscribers([topicname])
        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertTrue(topicname not in dt.added)  # not detected
        # make sure the topic is STILL in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has STILL not been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # create the publisher and then try updating again, simulating
        # it coming online after expose call.
        nonexistent_pub = rospy.Subscriber(topicname, Empty, queue_size=1)
        with Timeout(5) as t:
            while not t.timed_out and topicname not in dt.added:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.removed, [])  # nothing removed
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(topicname in dt.added)  # detected
        # TODO : do we need a test with subscriber ?

        # every exposed topic should remain in the list of args ( in case regex match another topic )
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has been created
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        # removing publisher
        nonexistent_pub.unregister()  # https://github.com/ros/ros_comm/issues/111 ( topic is still registered on master... )

        # and update should be enough to cleanup
        with Timeout(5) as t:
            while not t.timed_out and not topicname in dt.removed:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.added, [])  # nothing added
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(topicname in dt.removed)
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_available)

        # every exposed topic should still be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # the backend should not be there any longer
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

    def test_subscriber_withhold_update_disappear(self):
        """
        Test topic withholding functionality for a topic which doesnt exists anymore in the ros environment.
        Normal usecase
        Sequence : (-> UPDATE ?) -> WITHHOLD -> UPDATE -> DISAPPEAR (-> UPDATE ?)
        :return:
        """
        topicname = '/test/nonexistent3'
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        # First update should not change state
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # create the publisher and then try exposing the topic again, simulating
        # it coming online before expose call.
        nonexistent_pub = rospy.Subscriber(topicname, Empty, queue_size=1)
        with Timeout(5) as t:
            while not t.timed_out and topicname not in self.subscriber_if_pool.subscribers_available:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.added, [])  # nothing added (not exposed)
                self.assertEqual(dt.removed, [])  # nothing removed
                time.sleep(0.1)  # to avoid spinning out of control
        # TODO : do we need a test with subscriber ?

        self.assertTrue(not t.timed_out)
        # topic should be in the list of args yet (not exposed)
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # Here we are sure the internal state has topic_name registered
        # will be exposed right away
        dt = self.subscriber_if_pool.expose_subscribers([topicname])
        self.assertTrue(topicname in dt.added)  # detected and added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every exposed topic should remain in the list of args ( in case regex match another topic )
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has been created
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every withhold topic should STILL be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # topic backend should STILL be there
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        dt = self.subscriber_if_pool.expose_subscribers([])
        self.assertTrue(topicname in dt.removed)  # removed
        self.assertEqual(dt.added, [])  # nothing added
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend should be GONE
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend should be GONE
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        nonexistent_pub.unregister()

        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend should be GONE
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # Waiting a bit until the system state is back as expected (with connection cache this can be delayed)
        with Timeout(5) as t:
            while not t.timed_out and not topicname in self.subscriber_if_pool.subscribers_available:
                time.sleep(1)

    def test_subscriber_disappear_update_withhold(self):
        """
        Test topic exposing functionality for a topic which already exists in the ros environment.
        Normal usecase
        Sequence : (UPDATE? ->) DISAPPEAR -> UPDATE -> WITHHOLD (-> UPDATE ?)
        :return:
        """

        topicname = '/test/nonexistent4'
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        # First update should not change state
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        dt = self.subscriber_if_pool.expose_subscribers([topicname])
        self.assertEqual(dt.added, [])  # nothing added (topic name doesnt exist yet)
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # topic backend has been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        def appear_disappear():
            # create the publisher and then try exposing the topic again, simulating
            # it coming online before expose call.
            nonexistent_pub = rospy.Subscriber(topicname, Empty, queue_size=1)

            with Timeout(5) as t:
                dt = DiffTuple([], [])
                while not t.timed_out and nonexistent_pub.resolved_name not in dt.added:
                    subscribers, topic_types = self.get_system_state()
                    dt = self.subscriber_if_pool.update(subscribers, topic_types)
                    self.assertEqual(dt.removed, [])  # nothing removed
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(nonexistent_pub.resolved_name in dt.added)  # detected
            # TODO : do we need a test with subscriber ?

            # every added topic should be in the list of args
            self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
            # topic backend has been created
            self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

            # up to here possible sequences should have been already tested by previous tests
            # Now comes our actual disappearance / withholding test
            nonexistent_pub.unregister()

            # every added topic should be in the list of args
            self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
            # the backend should STILL be there
            self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())
            # Note the Topic implementation should take care of possible errors in this case

            with Timeout(5) as t:
                dt = DiffTuple([], [])
                while not t.timed_out and topicname not in dt.removed:
                    subscribers, topic_types = self.get_system_state()
                    dt = self.subscriber_if_pool.update(subscribers, topic_types)
                    self.assertEqual(dt.added, [])  # nothing added
                    time.sleep(0.1)  # to avoid spinning out of control

            self.assertTrue(not t.timed_out)
            self.assertTrue(topicname in dt.removed)  # detected lost
            # every exposed topic should remain in the list of args ( in case regex match another topic )
            self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
            # make sure the topic backend should NOT be there any longer
            self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        appear_disappear()

        # test that coming back actually also works

        # Note : for the cache version we need update to run in between to make sure the cache callback for dropped
        # interface doesnt interfere with the new topic coming up.
        # TODO fix this problem (cache node / interface update not spinning fast enough to detect the change)
        # HOW ?
        # -> cache node (or proxy??) with possibility to mask subscribers + node to not trigger change ???
        # -> splitting subscribers and publishers will make it better to not confuse what goes up/down ???
        # -> wait for deletion confirmation before diff reporting deleted ? in update loop or higher, at test/user level ?
        # Currently there are tricks in place to determine if a diff comes from interface pub/sub creation/deletion
        #
        time.sleep(2)
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)

        # test that coming back actually also works
        appear_disappear()

        self.subscriber_if_pool.expose_subscribers([])
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend has not been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertTrue(topicname not in dt.added)  # not detected
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # make sure the topic backend has been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

    def test_subscriber_update_disappear_withhold(self):
        """
        Test topic exposing functionality for a topic which already exists in
        the ros environment. Simple Normal usecase
        Sequence : UPDATE -> DISAPPEAR -> WITHHOLD
        :return:
        """

        topicname = '/test/nonexistent5'
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())
        # First update should not change state
        subscribers, topic_types = self.get_system_state()
        dt = self.subscriber_if_pool.update(subscribers, topic_types)
        self.assertEqual(dt.added, [])  # nothing added
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # the backend should not have been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        dt = self.subscriber_if_pool.expose_subscribers([topicname])
        self.assertEqual(dt.added, [])  # nothing added yet ( not existing )
        self.assertEqual(dt.removed, [])  # nothing removed
        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # topic backend has not been created
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())

        # create the publisher and then try exposing the topic again, simulating
        # it coming online before expose call.
        nonexistent_pub = rospy.Subscriber(topicname, Empty, queue_size=1)

        with Timeout(5) as t:
            dt = DiffTuple([], [])
            while not t.timed_out and nonexistent_pub.resolved_name not in dt.added:
                subscribers, topic_types = self.get_system_state()
                dt = self.subscriber_if_pool.update(subscribers, topic_types)
                self.assertEqual(dt.removed, [])  # nothing removed
                time.sleep(0.1)  # to avoid spinning out of control

        self.assertTrue(not t.timed_out)
        self.assertTrue(nonexistent_pub.resolved_name in dt.added)  # added now because it just appeared
        self.assertEqual(dt.removed, [])  # nothing removed
        # TODO : do we need a test with subscriber ?

        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # topic backend has been created
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())

        # up to here possible sequences should have been already tested by previous tests
        # Now comes our actual disappearrence / withholding test

        nonexistent_pub.unregister()  # https://github.com/ros/ros_comm/issues/111 ( topic is still registered on master... )
        # TODO : test disappear ( how ? )

        # every added topic should be in the list of args
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers_args)
        # the backend should STILL be there
        self.assertTrue(topicname in self.subscriber_if_pool.subscribers.keys())
        # Note the Topic implementation should take care of possible errors in this case

        self.subscriber_if_pool.expose_subscribers([])
        # every withhold topic should NOT be in the list of args
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers_args)
        # topic backend should NOT be there any longer
        self.assertTrue(topicname not in self.subscriber_if_pool.subscribers.keys())


#TODO : here we always test the full update => test the diff algorithm as well !

@nose.tools.istest
class TestRosInterface1NoCache(TestRosSubscriberIfPool):

    @nose.tools.nottest
    def get_system_state(self):

        publishers, subscribers, services = self._master.getSystemState()[2]
        topic_types = self._master.getTopicTypes()[2]

        return subscribers, topic_types

    def setUp(self):
        self.strsub = rospy.Subscriber('/test/string', String, queue_size=1)
        self.empsub = rospy.Subscriber('/test/empty', Empty, queue_size=1)

        self._master = rospy.get_master()
        self.subscriber_if_pool = RosSubscriberIfPool()

    def tearDown(self):
        self.subscriber_if_pool = None


# Testing with Connection Cache
@nose.tools.istest
class TestRosInterfaceCache(TestRosSubscriberIfPool):

    @nose.tools.nottest
    def get_system_state(self):
        # Note getting the system state via this interface, there is no need for callback
        publishers, subscribers, services = self.connection_cache.getSystemState()
        topic_types = self.connection_cache.getTopicTypes()

        return subscribers, topic_types

    def setUp(self):

        # first we setup our publishers and our node (used by rospy.resolve_name calls to remap topics)
        self.strsub = rospy.Subscriber('/test/string', String, queue_size=1)
        self.empsub = rospy.Subscriber('/test/empty', Empty, queue_size=1)

        self._master = rospy.get_master()
        self.subscriber_if_pool = RosSubscriberIfPool()

        # we need to speed fast enough for the tests to not fail on timeout...
        rospy.set_param('/connection_cache/spin_freq', 2)  # 2 Hz
        self.connection_cache_node = roslaunch.core.Node('rocon_python_comms', 'connection_cache.py',
                                                         name='connection_cache',
                                                         remap_args=[('~list', rospy.resolve_name('~connections_list')),
                                                                     (
                                                                     '~diff', rospy.resolve_name('~connections_diff'))])

        # Easier to remap the node topic to the proxy ones, instead of the opposite, since there is no dynamic remapping.
        # However for normal usecase, remapping the proxy handles is preferable.
        try:
            self.connection_cache_proc = self.launch.launch(self.connection_cache_node)
        except roslaunch.RLException as rlexc:
            raise nose.SkipTest("Connection Cache Node not found (part of rocon_python_comms pkg). Skipping test.")

        assert self.connection_cache_proc.is_alive()

        # wait for node to be started
        node_api = None
        with Timeout(5) as t:
            while not t.timed_out and node_api is None:
                node_api = rosnode.get_api_uri(rospy.get_master(), 'connection_cache')

        assert node_api is not None  # make sure the connection cache node is started before moving on.

        self.connection_cache = connection_cache_proxy_create()

    def tearDown(self):
        self.subscriber_if_pool = None

        self.connection_cache_proc.stop()
        while self.connection_cache_proc.is_alive():
            time.sleep(0.2)  # waiting for cache node to die
        assert not self.connection_cache_proc.is_alive()
        time.sleep(1)  # TODO : investigate : we shouldnt need this


if __name__ == '__main__':

    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('test_subscriber_if_pool_no_cache', 'test_all', TestRosInterface1NoCache)

    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('test_subscriber_if_pool_cache', 'test_all', TestRosInterfaceCache)



from __future__ import absolute_import
from __future__ import print_function

from collections import OrderedDict

from .api import rospy_safe as rospy
from ..baseinterface import TransientIf


class ParamTuple(object):
    def __init__(self, name, type):
        self.name = name
        self.type = type
# TODO: make that the pickled representation of ParamBack (check asdict())


class ParamBack(TransientIf):
    """
    ParamBack is the class handling conversion from REST API to ROS Param
    """
    def __init__(self, param_name, param_type):
        param_name = rospy.resolve_name(param_name)

        # defining a type on param to unify API for topic, services and param.
        super(ParamBack, self).__init__(param_name, param_type)

        # TODO : should we here manage how to get param values with rospy (same as for service and topic ?)

    def cleanup(self):
        super(ParamBack, self).cleanup()

    def asdict(self):
        """
        Here we provide a dictionary suitable for a representation of the Topic instance
        the main point here is to make it possible to transfer this to remote processes.
        We are not interested in pickleing the whole class with Subscriber and Publisher
        :return:
        """

        return OrderedDict({
            'name': self.name,
            'fullname': self.name,  # for BWcompat
            'prmtype': self.type,
        })

    def setval(self, val):
        # TODO : think about stricter type checking...
        rospy.set_param(self.name, val)
        return

    def getval(self):
        res = rospy.get_param(self.name)
        return res



from __future__ import absolute_import

import time
from collections import deque

from .api import rospy_safe as rospy
from .message_conversion import extract_values, FieldTypeMismatchException
from .poolparam import PoolParam
from .topicbase import TopicBase


class PublisherBackTimeout(Exception):
    pass


class PublisherBack(TopicBase):
    """
    TopicBack is the class handling conversion from Python to ROS Topic
    Requirement : Only one topicBack per actual ROS Topic.
    Since we connect to an already existing ros topic, our number of connections should never drop under 1
    """

    # We need some kind of instance count here since system state returns only one node instance
    # as publisher for multiple publisher in this process.
    #
    # This is used in ros_interface update for added puba/subs to determine
    # if we previously started publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests
    #
    # This is also used in ros_interface update for removed pubs/subs to determine
    # if we previously stopped publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests

    pool = PoolParam(rospy.Subscriber, "subscribers")

    def __init__(self, topic_name, topic_type, msg_queue_size=1):
        # Parent class will resolve/normalize topic_name
        super(PublisherBack, self).__init__(topic_name, topic_type)

        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Adding subscriber interface {name} {typename}".format(
                name=self.name, typename=self.rostype))

        # this message queue should be ready before we setup the callback
        # TODO : change to a proper Queue
        self.msg = deque([], msg_queue_size)

        self.topic = self.pool.acquire(self.name, self.rostype, self.topic_callback, queue_size=1)

        self.empty_cb = None

        # Make sure we get at least one connection before returning
        start = time.time()
        timeout = 1
        while time.time() - start < timeout and len(self.pool.get_impl_connections(self.name)) < 1:
            # print("subscribers connected : {0}".format(self.pool.get_impl_connections(self.name)))
            time.sleep(0.2)

        if self.pool.get_impl_connections(self.name) < 1:
            raise PublisherBackTimeout

    def cleanup(self):
        """
        Launched when we want to whithhold this interface instance
        :return:
        """

        # TODO : should we do this in del method instead ? to allow reuse until garbage collection actually happens...
        rospy.loginfo(
            rospy.get_name() + " Pyros.rosinterface : Removing subscriber interface {name} {typename}".format(
                name=self.name, typename=self.rostype))

        self.pool.release(self.topic)

        super(PublisherBack, self).cleanup()

    def asdict(self):
        """
        Here we provide a dictionary suitable for a representation of the Topic instance
        the main point here is to make it possible to transfer this to remote processes.
        We are not interested in pickleing the whole class with Subscriber and Publisher
        :return:
        """
        d = super(PublisherBack, self).asdict()
        d['publishers'] = self.topic.impl.get_stats_info()
        return d

    def get(self, num=0, consume=False):
        if not self.msg:
            return None
        # TODO : implement a way to have "plug and play" behaviors (some can be "all, paged, FIFO, etc." with custom code that can be insterted here...)
        res = None
        #TODO : implement returning multiple messages ( paging/offset like for long REST requests )
        if consume:
            res = self.msg.popleft()
            if 0 == len(self.msg) and self.empty_cb:
                self.empty_cb()
                #TODO : CHECK that we can survive here even if we get dropped from the topic list
        else:
            res = self.msg[0]
            try:
                res = extract_values(res)
            except FieldTypeMismatchException as e:
                rospy.logerr("[{name}] : field type mismatch {e}".format(name=__name__, e=e))
                raise
                # TODO : reraise a topic exception ?
        return res

    #returns the number of unread message
    def unread(self):
        return len(self.msg)

    def set_empty_callback(self, cb):
        self.empty_cb = cb

    def topic_callback(self, msg):
        # TODO : we are duplicating the queue behavior that is already in rospy... Is there a better way ?
        self.msg.appendleft(msg)

